# Blackbox tests for ndrdump
# Copyright (C) 2008 Andrew Tridgell <tridge@samba.org>
# Copyright (C) 2008 Andrew Bartlett <abartlet@samba.org>
# Copyright (C) 2010 Jelmer Vernooij <jelmer@samba.org>
# based on test_smbclient.sh

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Blackbox tests for ndrdump."""

import os
import re
from samba.tests import BlackboxTestCase, BlackboxProcessError

data_path_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "../../../../../source4/librpc/tests"))

class NdrDumpTests(BlackboxTestCase):
    """Blackbox tests for ndrdump."""

    def data_path(self, name):
        return os.path.join(data_path_dir, name)

    def test_ndrdump_with_in(self):
        self.check_run(("ndrdump --debug-stdout samr samr_CreateUser in %s" %
                       (self.data_path("samr-CreateUser-in.dat"))))

    def test_ndrdump_with_out(self):
        self.check_run(("ndrdump --debug-stdout samr samr_CreateUser out %s" %
                       (self.data_path("samr-CreateUser-out.dat"))))

    def test_ndrdump_context_file(self):
        self.check_run(
            ("ndrdump --debug-stdout --context-file %s samr samr_CreateUser out %s" %
                (self.data_path("samr-CreateUser-in.dat"),
                self.data_path("samr-CreateUser-out.dat"))))

    def test_ndrdump_with_validate(self):
        self.check_run(("ndrdump --debug-stdout --validate samr samr_CreateUser in %s" %
                       (self.data_path("samr-CreateUser-in.dat"))))

    def test_ndrdump_with_hex_decode_function(self):
        self.check_run(
            ("ndrdump --debug-stdout dns decode_dns_name_packet in --hex-input %s" %
                self.data_path("dns-decode_dns_name_packet-hex.dat")))

    def test_ndrdump_with_hex_struct_name(self):
        expected = open(self.data_path("dns-decode_dns_name_packet-hex.txt")).read()
        try:
            actual = self.check_output(
                "ndrdump --debug-stdout dns dns_name_packet struct --hex-input %s" %
                self.data_path("dns-decode_dns_name_packet-hex.dat"))
        except BlackboxProcessError as e:
            self.fail(e)

        # check_output will return bytes
        # convert expected to bytes for python 3
        self.assertEqual(actual, expected.encode('utf-8'))

    def test_ndrdump_with_binary_struct_name(self):
        # Prefix of the expected unparsed PAC data (without times, as
        # these vary by host)
        expected = '''pull returned Success
    PAC_DATA: struct PAC_DATA
        num_buffers              : 0x00000005 (5)
        version                  : 0x00000000 (0)
        buffers: ARRAY(5)'''
        try:
            actual = self.check_output(
                "ndrdump --debug-stdout krb5pac PAC_DATA struct %s" %
                self.data_path("krb5pac-PAC_DATA.dat"))
        except BlackboxProcessError as e:
            self.fail(e)

        # check_output will return bytes
        # convert expected to bytes for python 3
        self.assertEqual(actual[:len(expected)],
                         expected.encode('utf-8'))
        self.assertTrue(actual.endswith(b"dump OK\n"))

    def test_ndrdump_upn_dns_info_ex(self):
        with open(self.data_path(
                'krb5pac_upn_dns_info_ex.txt')) as f:
            expected = f.read()
        data_path = self.data_path(
            'krb5pac_upn_dns_info_ex.b64.txt')

        try:
            actual = self.check_output(
                'ndrdump --debug-stdout -d0 krb5pac PAC_DATA struct '
                '--validate --base64-input ' + data_path)
        except BlackboxProcessError as e:
            self.fail(e)

        self.assertEqual(actual, expected.encode('utf-8'))

    def test_ndrdump_upn_dns_info_ex_not_supported(self):
        with open(self.data_path(
                'krb5pac_upn_dns_info_ex_not_supported.txt')) as f:
            expected = f.read()
        data_path = self.data_path(
            'krb5pac_upn_dns_info_ex_not_supported.b64.txt')

        try:
            # This PAC has been edited to remove the
            # PAC_UPN_DNS_FLAG_HAS_SAM_NAME_AND_SID bit, so that we can
            # simulate older versions of Samba parsing this structure.
            actual = self.check_output(
                'ndrdump --debug-stdout -d0 krb5pac PAC_DATA struct '
                '--validate --base64-input ' + data_path)
        except BlackboxProcessError as e:
            self.fail(e)

        self.assertEqual(actual, expected.encode('utf-8'))

    def test_ndrdump_with_binary_struct_number(self):
        expected = '''pull returned Success
    GUID                     : 33323130-3534-3736-3839-616263646566
dump OK
'''
        try:
            actual = self.check_output(
                "ndrdump --debug-stdout misc 0 struct %s" %
                self.data_path("misc-GUID.dat"))
        except BlackboxProcessError as e:
            self.fail(e)

        # check_output will return bytes
        # convert expected to bytes for python 3
        self.assertEqual(actual, expected.encode('utf-8'))

    def test_ndrdump_with_enum_not_struct(self):
        expected = '''Public structure 'netr_SchannelType' not found
'''
        try:
            actual = self.check_exit_code(
                "ndrdump --debug-stdout misc netr_SchannelType --input=x struct",
                1)
        except BlackboxProcessError as e:
            self.fail(e)

        # check_output will return bytes
        # convert expected to bytes for python 3
        self.assertEqual(actual, expected.encode('utf-8'))

    def test_ndrdump_input_cmdline_short_struct_name(self):
        expected = '''pull returned Buffer Size Error
'''
        try:
            actual = self.check_exit_code(
                "ndrdump --debug-stdout -d0 misc GUID struct --input=abcdefg", 2)
        except BlackboxProcessError as e:
            self.fail(e)

        # check_output will return bytes
        # convert expected to bytes for python 3
        self.assertEqual(actual, expected.encode('utf-8'))

    def test_ndrdump_input_cmdline_short_struct_name_dump(self):
        expected = '''pull returned Buffer Size Error
6 bytes consumed
[0000] 61 62 63 64 65 66 67                                abcdefg''' \
        '''
'''
        try:
            actual = self.check_exit_code(
                "ndrdump --debug-stdout -d0 misc GUID struct --input=abcdefg --dump-data", 2)
        except BlackboxProcessError as e:
            self.fail(e)

        # check_output will return bytes
        # convert expected to bytes for python 3
        self.assertEqual(actual, expected.encode('utf-8'))

    def test_ndrdump_input_cmdline_short_struct_name_print_fail(self):
        expected = '''pull returned Buffer Size Error
6 bytes consumed
[0000] 61 62 63 64 65 66 67                                abcdefg''' \
        '''
WARNING! 1 unread bytes
[0000] 67                                                  g''' \
    '''
WARNING: pull of GUID was incomplete, therefore the parse below may SEGFAULT
    GUID                     : 64636261-6665-0000-0000-000000000000
dump of failed-to-parse GUID complete
'''
        try:
            actual = self.check_exit_code(
                "ndrdump --debug-stdout -d0 misc GUID struct --input=abcdefg --dump-data --print-after-parse-failure", 2)
        except BlackboxProcessError as e:
            self.fail(e)

        # check_output will return bytes
        # convert expected to bytes for python 3
        self.assertEqual(actual, expected.encode('utf-8'))

    def test_ndrdump_fuzzed_clusapi_QueryAllValues(self):
        expected = b'''pull returned Success
WARNING! 53 unread bytes
[0000] 00 FF 00 00 FF 00 00 00   00 09 00 00 00 08 00 33   ........ .......3
[0010] 33 32 37 36 32 36 39 33   32 37 36 38 34 01 00 00   32762693 27684...
[0020] 80 32 0D FF 00 00 FF 00   00 00 00 08 00 00 00 1C   .2...... ........
[0030] F1 29 08 00 00                                      .)...''' \
        b'''
    clusapi_QueryAllValues: struct clusapi_QueryAllValues
        out: struct clusapi_QueryAllValues
            pcbData                  : *
                pcbData                  : 0x01000000 (16777216)
            ppData                   : *
                ppData: ARRAY(1)
                    ppData                   : NULL
            rpc_status               : *
                rpc_status               : WERR_OK
            result                   : WERR_NOT_ENOUGH_MEMORY
dump OK
'''
        try:
            actual = self.check_output(
                'ndrdump --debug-stdout clusapi clusapi_QueryAllValues out ' +\
                '--base64-input --input=' +\
                'AAAAAQEAAAAAAAAAAAAAAAgAAAAA/wAA/wAAAAAJAAAACAAzMzI3NjI2OTMyNzY4NAEAAIAyDf8AAP8AAAAACAAAABzxKQgAAA==')
        except BlackboxProcessError as e:
            self.fail(e)
        self.assertEqual(actual, expected)

    def test_ndrdump_fuzzed_ntlmsssp_AUTHENTICATE_MESSAGE(self):
        expected = open(self.data_path("fuzzed_ntlmssp-AUTHENTICATE_MESSAGE.txt")).read()
        try:
            actual = self.check_output(
                "ndrdump --debug-stdout ntlmssp AUTHENTICATE_MESSAGE struct --base64-input %s --validate" %
                self.data_path("fuzzed_ntlmssp-AUTHENTICATE_MESSAGE.b64.txt"))
        except BlackboxProcessError as e:
            self.fail(e)
        # check_output will return bytes
        # convert expected to bytes for python 3
        self.assertEqual(actual, expected.encode('utf-8'))

    def test_ndrdump_fuzzed_PackagesBlob(self):
        expected = 'ndr_pull_string: ndr_pull_error\\(Buffer Size Error\\):'
        command = (
            "ndrdump --debug-stdout drsblobs package_PackagesBlob struct --input='aw=='"
            " --base64-input")
        try:
            actual = self.check_exit_code(command, 2)
        except BlackboxProcessError as e:
            self.fail(e)
        # check_output will return bytes
        # convert expected to bytes for python 3
        self.assertRegex(actual.decode('utf8'), expected)

    def test_ndrdump_fuzzed_drsuapi_DsAddEntry_1(self):
        expected = open(self.data_path("fuzzed_drsuapi_DsAddEntry_1.txt")).read()
        try:
            actual = self.check_output(
                "ndrdump --debug-stdout drsuapi drsuapi_DsAddEntry in --base64-input --validate %s" %
                self.data_path("fuzzed_drsuapi_DsAddEntry_1.b64.txt"))
        except BlackboxProcessError as e:
            self.fail(e)
        # check_output will return bytes
        # convert expected to bytes for python 3
        self.assertEqual(actual, expected.encode('utf-8'))

    def test_ndrdump_fuzzed_drsuapi_DsaAddressListItem_V1(self):
        expected = "Maximum Recursion Exceeded"
        try:
            self.check_output(
                "ndrdump --debug-stdout drsuapi 17 out --base64-input %s" %
                self.data_path(
                    "fuzzed_drsuapi_DsaAddressListItem_V1-in.b64.txt"))
            self.fail("Input should have been rejected with %s" % expected)
        except BlackboxProcessError as e:
            if expected not in str(e):
                self.fail(e)

    def test_ndrdump_fuzzed_drsuapi_DsReplicaAttribute(self):
        expected = open(self.data_path("fuzzed_drsuapi_DsReplicaAttribute.txt")).read()
        try:
            actual = self.check_output(
                "ndrdump --debug-stdout drsuapi drsuapi_DsReplicaAttribute struct --base64-input --validate %s" %
                self.data_path("fuzzed_drsuapi_DsReplicaAttribute.b64.txt"))
        except BlackboxProcessError as e:
            self.fail(e)
        # check_output will return bytes
        # convert expected to bytes for python 3
        self.assertEqual(actual, expected.encode('utf-8'))

    def test_ndrdump_Krb5ccache(self):
        expected = open(self.data_path("../../../source3/selftest/"
                                       "ktest-krb5_ccache-2.txt")).read()
        try:
            # Specify -d1 to match the generated output file, because ndrdump
            # only outputs some additional info if this parameter is specified,
            # and the --configfile parameter gives us an empty smb.conf to avoid
            # extraneous output.
            actual = self.check_output(
                "ndrdump krb5ccache CCACHE struct "
                "--configfile /dev/null --debug-stdout -d1 --validate " +
                self.data_path("../../../source3/selftest/"
                               "ktest-krb5_ccache-2"))
        except BlackboxProcessError as e:
            self.fail(e)
        # check_output will return bytes
        # convert expected to bytes for python 3
        self.assertEqual(actual, expected.encode('utf-8'))

        expected = open(self.data_path("../../../source3/selftest/"
                                       "ktest-krb5_ccache-3.txt")).read()
        try:
            # Specify -d1 to match the generated output file, because ndrdump
            # only outputs some additional info if this parameter is specified,
            # and the --configfile parameter gives us an empty smb.conf to avoid
            # extraneous output.
            actual = self.check_output(
                "ndrdump krb5ccache CCACHE struct "
                "--configfile /dev/null --debug-stdout -d1 --validate " +
                self.data_path("../../../source3/selftest/"
                               "ktest-krb5_ccache-3"))
        except BlackboxProcessError as e:
            self.fail(e)
        # check_output will return bytes
        # convert expected to bytes for python 3
        self.assertEqual(actual, expected.encode('utf-8'))

    # This is a good example of a union with an empty default
    # and no buffers to parse.
    def test_ndrdump_fuzzed_spoolss_EnumForms(self):
        expected_head = b'''pull returned Success
WARNING! 2 unread bytes
[0000] 00 00                                               ..''' b'''
    spoolss_EnumForms: struct spoolss_EnumForms
        out: struct spoolss_EnumForms
            count                    : *
                count                    : 0x00000100 (256)
            info                     : *
                info                     : *
                    info: ARRAY(256)
                        info                     : union spoolss_FormInfo(case 0)
                        info                     : union spoolss_FormInfo(case 0)
'''
        expected_tail = b'''info                     : union spoolss_FormInfo(case 0)
                        info                     : union spoolss_FormInfo(case 0)
                        info                     : union spoolss_FormInfo(case 0)
                        info                     : union spoolss_FormInfo(case 0)
                        info                     : union spoolss_FormInfo(case 0)
                        info                     : union spoolss_FormInfo(case 0)
            needed                   : *
                needed                   : 0x00000000 (0)
            result                   : HRES code 0xa9a9a900
dump OK
'''
        try:
            actual = self.check_output(
                "ndrdump --debug-stdout spoolss spoolss_EnumForms out --base64-input " +\
                "--input AAAAAQAAAAAAAAAAAAEAAACpqakAAA="
                )
        except BlackboxProcessError as e:
            self.fail(e)
        self.assertEqual(actual[:len(expected_head)],
                         expected_head)
        self.assertTrue(actual.endswith(expected_tail))

    # This is a good example of a union with scalars and buffers
    def test_ndrdump_xattr_NTACL(self):

        expected_head =  open(self.data_path("xattr_NTACL.txt")).read().encode('utf8')
        expected_tail = b'''dump OK
'''
        try:
            actual = self.check_output(
                "ndrdump --debug-stdout xattr xattr_NTACL struct --hex-input %s --validate" %
                self.data_path("xattr_NTACL.dat"))
        except BlackboxProcessError as e:
            self.fail(e)

        self.assertEqual(actual[:len(expected_head)],
                         expected_head)
        self.assertTrue(actual.endswith(expected_tail))

    # Confirm parsing of dnsProperty records
    def test_ndrdump_dnsp_DnssrvRpcRecord(self):

        expected = open(self.data_path("dnsp-DnssrvRpcRecord.txt")).read().encode('utf8')
        try:
            actual = self.check_output(
                "ndrdump --debug-stdout dnsp dnsp_DnssrvRpcRecord struct " +\
                "--input BQAPAAXwAAC3AAAAAAADhAAAAAAAAAAAAAoBAAA= "+\
                "--base64-input --validate")
        except BlackboxProcessError as e:
            self.fail(e)

        self.assertEqual(actual, expected)

    # Test a --validate push of a NULL union pointer
    def test_ndrdump_fuzzed_NULL_union_PAC_BUFFER(self):
        expected = b'''pull returned Success
WARNING! 13 unread bytes
[0000] F5 FF 00 3C 3C 25 FF 70   16 1F A0 12 84            ...<<%.p .....
    PAC_BUFFER: struct PAC_BUFFER
        type                     : UNKNOWN_ENUM_VALUE (1094251328)
        _ndr_size                : 0x048792c6 (75993798)
        info                     : NULL
        _pad                     : 0x06000000 (100663296)
push returned Success
pull returned Success
    PAC_BUFFER: struct PAC_BUFFER
        type                     : UNKNOWN_ENUM_VALUE (1094251328)
        _ndr_size                : 0x00000000 (0)
        info                     : NULL
        _pad                     : 0x00000000 (0)
WARNING! orig bytes:29 validated pushed bytes:16
WARNING! orig and validated differ at byte 0x04 (4)
WARNING! orig byte[0x04] = 0xC6 validated byte[0x04] = 0x00
-[0000] 40 F3 38 41 C6 92 87 04   00 00 00 00 00 00 00 06   @.8A.... ........
+[0000] 40 F3 38 41 00 00 00 00   00 00 00 00 00 00 00 00   @.8A.... ........
-[0010] F5 FF 00 3C 3C 25 FF 70   16 1F A0 12 84            ...<<%.p .....
+[0010]                                                     EMPTY   BLOCK
dump OK
'''
        try:
            actual = self.check_output(
                "ndrdump --debug-stdout krb5pac PAC_BUFFER struct --validate --input " +\
                "QPM4QcaShwQAAAAAAAAABvX/ADw8Jf9wFh+gEoQ= --base64-input")
        except BlackboxProcessError as e:
            self.fail(e)

        self.assertEqual(actual, expected)

    # Test a --validate push of a NULL struct pointer
    def test_ndrdump_fuzzed_NULL_struct_ntlmssp_CHALLENGE_MESSAGE(self):
        expected =  open(self.data_path("fuzzed_ntlmssp-CHALLENGE_MESSAGE.txt")).read().encode('utf8')
        try:
            actual = self.check_output(
                "ndrdump --debug-stdout ntlmssp CHALLENGE_MESSAGE struct --validate --input " +\
                "'AAAACwIAAAAAJwIAAAAAAAcAAAAAAAAAAIAbhG8uyk9dAL0mQE73MAAAAAAAAAAA' --base64-input")
        except BlackboxProcessError as e:
            self.fail(e)

        # Filter out the C source file and line number
        regex = rb"\.\./\.\./librpc/ndr/ndr\.c:[0-9]+"
        actual = re.sub(regex, b"", actual)
        expected = re.sub(regex, b"", expected)

        self.assertEqual(actual, expected)

    # Test a print of NULL pointer in manually-written ndr_drsuapi.c
    def test_fuzzed_drsuapi_DsGetNCChanges(self):
        expected =  open(self.data_path("fuzzed_drsuapi_DsGetNCChanges.txt"), 'rb').read()
        try:
            actual = self.check_output(
                "ndrdump --debug-stdout drsuapi 3 out --base64-input --input " +\
                "AQAAAAEAAAAGAKoAAAAGAKoGAAMAAQAAAAYAEwAAAAAAAAAA/wAAAAAAAAA/AAAAAAAAAAAAAAAAAAAAAABbAAAAAAAAAAAAAAkRAAABAAAAAQAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAPkAAAAAAAABAAD4BgATAAAAAAAAAAD/AAAAAAAAAD8AAAAAAAAAAAAAAAAAAAAAAFsAAAAAAAAAAAAABgAQAAEAAAABAAAAAQAAAAEAAAABAAAAAQAAAAMAAAABAAAACREAAAEAAAABAAAAAAAAAAYAEAABAAgAAQAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAEAAAAAAAA=")
        except BlackboxProcessError as e:
            self.fail(e)

        self.assertEqual(actual, expected)

    def test_ndrdump_fuzzed_ndr_compression(self):
        expected = r'ndr_pull_compression_start: ndr_pull_error\(Compression Error\): Bad compression algorithm 204 \(PULL\)'
        command = (
            "ndrdump --debug-stdout drsuapi 3 out --base64-input "
            "--input BwAAAAcAAAAGAAAAAwAgICAgICAJAAAAICAgIAkAAAAgIAAA//////8=")
        try:
            actual = self.check_exit_code(command, 2)
        except BlackboxProcessError as e:
            self.fail(e)
        # check_output will return bytes
        # convert expected to bytes for python 3
        self.assertRegex(actual.decode('utf8'), expected)

    def test_ndrdump_short_dnsProperty(self):
        expected = b'''pull returned Success
    dnsp_DnsProperty_short: struct dnsp_DnsProperty_short
        wDataLength              : 0x00000000 (0)
        namelength               : 0x00000000 (0)
        flag                     : 0x00000000 (0)
        version                  : 0x00000001 (1)
        id                       : DSPROPERTY_ZONE_NS_SERVERS_DA (146)
        data                     : union dnsPropertyData(case 0)
        name                     : 0x00000000 (0)
dump OK
'''
        command = (
            "ndrdump --debug-stdout dnsp dnsp_DnsProperty_short struct --base64-input "
            "--input AAAAAAAAAAAAAAAAAQAAAJIAAAAAAAAA")
        try:
            actual = self.check_output(command)
        except BlackboxProcessError as e:
            self.fail(e)
        self.assertEqual(actual, expected)

    # This is compressed with Microsoft's compression, so we can't do a validate
    def test_ndrdump_compressed_claims(self):
        expected =  open(self.data_path("compressed_claims.txt"), 'rb').read()

        try:
            actual = self.check_output(
                "ndrdump --debug-stdout claims CLAIMS_SET_METADATA_NDR struct --hex-input --input " + \
                "01100800cccccccc500200000000000000000200290200000400020004000000282000000000000000000000000000002902000073778788878808880700080007800800060007000700070887770780080088008870070008000808000080000000008070787787770076770867868788000000000000000000000000000000000000000000000000000000000000000000000000000000000008000000000000000000000000000000000000000000000000007700080080000000870000000000000085070000000000007476800000000000750587000800000066078000000080706677880080008060878708000000008000800000000000800000000000000000000000000000000000000000000000006080080000000070000000000000000000000000000000000000000000000000fd74eaf001add6213aecf4346587eec48c323e3e1a5a32042eecf243669a581e383d2940e80e383c294463b8c0b49024f1def20df819586b086cd2ab98700923386674845663ef57e91718110c1ad4c0ac88912126d2180545e98670ea2aa002052aa54189cc318d26c46b667f18b6876262a9a4985ecdf76e5161033fd457ba020075360c837aaa3aa82749ee8152420999b553c60195be5e5c35c4330557538772972a7d527aeca1fc6b2951ca254ac83960272a930f3194892d4729eff48e48ccfb929329ff501c356c0e8ed18471ec70986c31da86a8090b4022c1db257514fdba4347532146648d4f99f9065e0d9a0d90d80f38389c39cb9ebe6d4e5e681e5a8a5418f591f1dbb7594a3f2aa3220ced1cd18cb49cffcc2ff18eef6caf443663640c5664000012000000")
        except BlackboxProcessError as e:
            self.fail(e)

        self.assertEqual(actual, expected)

    def test_ndrdump_uncompressed_claims(self):
        expected =  open(self.data_path("uncompressed_claims.txt"), 'rb').read()

        try:
            actual = self.check_output(
                "ndrdump --debug-stdout claims CLAIMS_SET_METADATA_NDR struct --hex-input --input " + \
                "01100800cccccccc800100000000000000000200580100000400020000000000580100000000000000000000000000005801000001100800cccccccc480100000000000000000200010000000400020000000000000000000000000001000000010000000300000008000200030000000c000200060006000100000010000200140002000300030003000000180002002800020002000200040000002c0002000b000000000000000b000000370032003000660064003300630033005f00390000000000010000000000000001000000000000000b000000000000000b000000370032003000660064003300630033005f00370000000000030000001c000200200002002400020004000000000000000400000066006f006f0000000400000000000000040000006200610072000000040000000000000004000000620061007a0000000b000000000000000b000000370032003000660064003300630033005f003800000000000400000009000a000000000007000100000000000600010000000000000001000000000000000000")
        except BlackboxProcessError as e:
            self.fail(e)

        self.assertEqual(actual, expected)

    # We can't run --validate here as currently we can't round-trip
    # this data due to uninitialised padding in the sample
    def test_ndrdump_claims_CLAIMS_SET_NDR(self):
        expected =  open(self.data_path("claims_CLAIMS_SET_NDR.txt"), 'rb').read()

        try:
            actual = self.check_output(
                "ndrdump --debug-stdout claims CLAIMS_SET_NDR struct --hex-input " + \
                self.data_path("claims_CLAIMS_SET_NDR.dat"))
        except BlackboxProcessError as e:
            self.fail(e)

        self.assertEqual(actual, expected)

    def test_ndrdump_gmsa_MANAGEDPASSWORD_BLOB(self):
        with open(self.data_path('gmsa_MANAGEDPASSWORD_BLOB.txt'), 'rb') as f:
            expected = f.read()

        try:
            actual = self.check_output(
                'ndrdump gmsa MANAGEDPASSWORD_BLOB struct '
                '--debug-stdout --validate --hex-input --input '
                '01000000220100001000000012011a01678657a1136e547f46ee7988c808'
                'd904ed0e4b0592f89eb82bd292685867c3119dd6eaaef5810a1aa4e08e49'
                '7cc31163b2e799e6ea66e3022c100bf59585a3464274ebad2488fc28acbd'
                '10a9b44dde436a6d35fff0e95ae7903609e825220ad30db6a86bb544fa34'
                '0f864d2d3895193d4007df72478d71ce3f789bb139c4c1cffd6d39948c0a'
                'fa6a65e3f5f8f90d8c70f7272ce65a3f632793eb0e4697e576c21f36ac55'
                'f4167a22b4ebb2593c2d22dc4ac8d4ca455f299a182b8d4d8dd1232dde1e'
                'fe3acaf14b137453195f45455f5d48a0c441913b80f94d4696b171379b5a'
                'c3b02c501cf8e16b43beaca52263411d5cf772e763e8d29a70a1293e7218'
                'a1e300007495016980170000743731b67f170000'
            )
        except BlackboxProcessError as e:
            self.fail(e)

        self.assertEqual(actual, expected)

    def test_ndrdump_keyetab(self):
        with open(self.data_path(
                'secrets.keytab.txt')) as f:
            expected = f.read()
        data_path = self.data_path(
            'secrets.keytab')

        # We don't do a validate on this as the parser deliberately
        # ignores some extra data written by Heimdal, but not MIT.
        try:
            actual = self.check_output(
                'ndrdump krb5ccache KEYTAB struct '
                + data_path)
        except BlackboxProcessError as e:
            self.fail(e)

        self.assertEqual(actual, expected.encode('utf-8'))



# Unix SMB/CIFS implementation.
# Copyright (C) Andrew Bartlett <abartlet@samba.org> 2018
# Written by Joe Guo <joeg@catalyst.net.nz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
from samba.tests.samba_tool.base import SambaToolCmdTest


class DemoteCmdTestCase(SambaToolCmdTest):
    """Test for samba-tool domain demote subcommand"""

    def setUp(self):
        super().setUp()
        self.creds_string = "-U{0}%{1}".format(
            os.environ["DC_USERNAME"], os.environ["DC_PASSWORD"])

        self.dc_server = os.environ['DC_SERVER']
        self.dburl = "ldap://%s" % os.environ["DC_SERVER"]
        self.samdb = self.getSamDB("-H", self.dburl, self.creds_string)

    def test_demote_and_remove_dns(self):
        """
        Test domain demote command will also remove dns references
        """

        server = os.environ['SERVER']  # the server to demote
        zone = os.environ['REALM'].lower()

        # make sure zone exist
        result, out, err = self.runsubcmd(
            "dns", "zoneinfo", server, zone, self.creds_string)
        self.assertCmdSuccess(result, out, err)

        # add a A record for the server to demote
        result, out, err = self.runsubcmd(
            "dns", "add", self.dc_server, zone,
            server, "A", "192.168.0.193", self.creds_string)
        self.assertCmdSuccess(result, out, err)

        # make sure above A record exist
        result, out, err = self.runsubcmd(
            "dns", "query", self.dc_server, zone,
            server, 'A', self.creds_string)
        self.assertCmdSuccess(result, out, err)

        # the above A record points to this host
        dnshostname = '{0}.{1}'.format(server, zone)

        # add a SRV record points to above host
        srv_record = "{0} 65530 65530 65530".format(dnshostname)
        self.runsubcmd(
            "dns", "add", self.dc_server, zone, 'testrecord', "SRV",
            srv_record, self.creds_string)

        # make sure above SRV record exist
        result, out, err = self.runsubcmd(
            "dns", "query", self.dc_server, zone,
            "testrecord", 'SRV', self.creds_string)
        self.assertCmdSuccess(result, out, err)

        for type_ in ['CNAME', 'NS', 'PTR']:
            # create record
            self.runsubcmd(
                "dns", "add", self.dc_server, zone,
                'testrecord', type_, dnshostname,
                self.creds_string)
            self.assertCmdSuccess(result, out, err)

            # check exist
            result, out, err = self.runsubcmd(
                "dns", "query", self.dc_server, zone,
                "testrecord", 'SRV', self.creds_string)
            self.assertCmdSuccess(result, out, err)

        # now demote
        result, out, err = self.runsubcmd(
            "domain", "demote",
            "--server", self.dc_server,
            "--configfile", os.environ["CONFIGFILE"],
            "--workgroup", os.environ["DOMAIN"],
            self.creds_string)
        self.assertCmdSuccess(result, out, err)

        result, out, err = self.runsubcmd(
            "dns", "query", self.dc_server, zone,
            server, 'ALL', self.creds_string)
        self.assertCmdFail(result)

        result, out, err = self.runsubcmd(
            "dns", "query", self.dc_server, zone,
            "testrecord", 'ALL', self.creds_string)
        self.assertCmdFail(result)



# Unix SMB/CIFS implementation.
#
# Tests for `samba-tool user keytrust`
#
# Copyright © Douglas Bagnall <dbagnall@samba.org> 2025
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
from pathlib import Path

from samba.domain.models import User
from samba.tests.samba_tool.base import SambaToolCmdTest
from samba import key_credential_link as kcl


HOST = "ldap://{DC_SERVER}".format(**os.environ)
CREDS = "-U{DC_USERNAME}%{DC_PASSWORD}".format(**os.environ)

ROOT = (Path(__file__) / '../../../../../').resolve()
TESTDATA = ROOT / 'testdata' / 'keytrust'

GOOD_CERTS = [
    str(TESTDATA / 'cert-rsa-2048.pem'),
    str(TESTDATA / 'ca-cert-rsa-2048.pem'),
]

WRONG_SIZE_CERTS = [
    str(TESTDATA / 'cert-rsa-1024.pem'),
    str(TESTDATA / 'ca-cert-rsa-4096.pem'),
]

NON_RSA_CERTS = [
    str(TESTDATA / 'ca-cert-ecdsa-p256.pem'),
]

GOOD_KEYS = [
    str(TESTDATA / 'rsa2048-pkcs1.der'),
    str(TESTDATA / 'rsa2048b-spki.pem'),
]

DUPLICATE_KEYS = [
    str(TESTDATA / 'cert-rsa-2048.pem'),
    str(TESTDATA / 'public-key-from-cert-rsa-2048-pkcs1.pem'),
]


class SambaToolUserKeyTrustTest(SambaToolCmdTest):
    cmd = 'user'
    user1 = 'alice'
    user2 = 'joe'
    ou_name = 'CN=Users'

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.samdb = cls.getSamDB("-H", HOST, CREDS)
        cls.runcmd(cls.cmd, "keytrust", "delete",
                   "-H", HOST, CREDS,
                   cls.user1, '--all')
        cls.runcmd(cls.cmd, "keytrust", "delete",
                   "-H", HOST, CREDS,
                   cls.user2, '--all')
        cls.ou = f"{cls.ou_name},{cls.samdb.domain_dn()}"

    def get_links(self, username):
        result = self.samdb.search(expression=f'sAMAccountName={username}',
                                   attrs=['msDS-KeyCredentialLink'])
        self.assertEqual(len(result), 1)
        links = result[0].get('msDS-KeyCredentialLink', [])
        return [kcl.KeyCredentialLinkDn(self.samdb, v) for v in links]

    def test_add_good_cert(self):
        """These ones should just succeed."""
        links = self.get_links(self.user2)
        n = len(links)
        for f in GOOD_CERTS:
            result, out, err = self.runcmd(self.cmd, "keytrust", "add",
                                           "-H", HOST, CREDS,
                                           self.user2, f)
            self.assertCmdSuccess(result, out, err)

            n += 1
            links = self.get_links(self.user2)
            self.assertEqual(len(links), n)

        result, out, err = self.runcmd(self.cmd, "keytrust", "delete",
                                       "-H", HOST, CREDS,
                                       self.user2, '--all')
        self.assertCmdSuccess(result, out, err)

        for link in links:
            self.assertIn(f"Deleted {link}", out)

        links = self.get_links(self.user2)
        self.assertEqual(links, [])

    def test_add_and_delete_good_keys(self):
        """Add known good keys, and also check the view and delete commands."""
        links = self.get_links(self.user1)
        self.assertEqual(links, [])

        result, out, err = self.runcmd(self.cmd, "keytrust", "add",
                                       "-H", HOST, CREDS,
                                       self.user1, GOOD_KEYS[0])
        self.assertCmdSuccess(result, out, err)
        links = self.get_links(self.user1)
        self.assertEqual(len(links), 1)

        result, out, err = self.runcmd(self.cmd, "keytrust", "view",
                                       "-H", HOST, CREDS,
                                       self.user1)
        self.assertCmdSuccess(result, out, err)
        self.assertIn(f'{self.user1} has 1 key credential link\n', out)
        self.assertIn(f'Link target: CN={self.user1},{self.ou}\n', out)
        self.assertIn('Number of key entries:            5', out)

        result, out, err = self.runcmd(self.cmd, "keytrust", "add",
                                       "-H", HOST, CREDS,
                                       self.user1, GOOD_KEYS[1])
        self.assertCmdSuccess(result, out, err)
        result, out, err = self.runcmd(self.cmd, "keytrust", "view",
                                       "-H", HOST, CREDS,
                                       self.user1, '--verbose')
        self.assertCmdSuccess(result, out, err)
        self.assertIn(f'{self.user1} has 2 key credential links\n', out)

        links = self.get_links(self.user1)
        fingerprints = [('16:CD:1B:C2:7A:0B:FC:C9:4B:95:11:9F:AD:97:EC:1B:'
                         'ED:BD:64:91:42:2E:AF:CA:CB:1E:C3:EE:86:6D:F1:5A'),
                        ('86:61:6D:B2:6A:3A:04:BD:E0:59:10:13:21:9A:2B:2C:'
                         'C4:FD:CE:50:05:16:3C:66:1B:38:63:79:8C:B1:DA:17')]

        self.assertEqual(set(x.fingerprint() for x in links),
                         set(fingerprints))

        # test delete --dry-run / -n
        result, out, err = self.runcmd(self.cmd, "keytrust", "delete",
                                       "-H", HOST, CREDS,
                                       self.user1, '--all', '-n')
        self.assertCmdSuccess(result, out, err)
        self.assertIn('Without --dry-run, this would happen:\n', out)
        self.assertIn(f'DELETE {links[0]} (fingerprint {links[0].fingerprint()})',
                      out)
        self.assertIn(f'DELETE {links[1]} (fingerprint {links[1].fingerprint()})',
                      out)
        self.assertNotIn('KEEP', out)
        self.assertIn(f'{self.user1} would now have 0 key credential links\n', out)

        result, out, err = self.runcmd(self.cmd, "keytrust", "delete",
                                       "-H", HOST, CREDS,
                                       self.user1, '--fingerprint=whatever',
                                       '--dry-run')
        self.assertCmdSuccess(result, out, err)
        self.assertIn('NO key credential links are deleted\n', out)

        self.assertIn(f'KEEP {links[0]} (fingerprint {links[0].fingerprint()})',
                      out)
        self.assertIn(f'KEEP {links[1]} (fingerprint {links[1].fingerprint()})',
                      out)
        self.assertIn(f'{self.user1} would now have 2 key credential links\n', out)

        result, out, err = self.runcmd(self.cmd, "keytrust", "delete",
                                       "-H", HOST, CREDS,
                                       self.user1,
                                       '--fingerprint',
                                       fingerprints[1],
                                       '--dry-run')
        self.assertCmdSuccess(result, out, err)
        self.assertIn(f'DELETE {links[1]} (fingerprint {links[1].fingerprint()})',
                      out)
        self.assertIn(f'KEEP {links[0]} (fingerprint {links[0].fingerprint()})',
                      out)
        self.assertIn(f'{self.user1} would now have 1 key credential link\n', out)

        # this time deleting for real
        result, out, err = self.runcmd(self.cmd, "keytrust", "delete",
                                       "-H", HOST, CREDS,
                                       self.user1, '--all')
        self.assertCmdSuccess(result, out, err)
        links = self.get_links(self.user1)
        self.assertEqual(links, [])

        result, out, err = self.runcmd(self.cmd, "keytrust", "view",
                                       "-H", HOST, CREDS,
                                       self.user1)
        self.assertCmdSuccess(result, out, err)
        self.assertIn(f'{self.user1} has 0 key credential links\n', out)

    def test_add_duplicate_keys(self):
        """You should not be able to add the same link twice."""

        self.addCleanup(self.runcmd, self.cmd, "keytrust", "delete",
                        "-H", HOST, CREDS,
                        self.user1, '--all')

        result, out, err = self.runcmd(self.cmd, "keytrust", "add",
                                       "-H", HOST, CREDS,
                                       self.user1, DUPLICATE_KEYS[0])
        self.assertCmdSuccess(result, out, err)

        # This source file is different, but contains the same public
        # key. samba-tool should notice this and fail *before* it
        # fails in the dsdb layer with ERR_ATTRIBUTE_OR_VALUE_EXISTS
        result, out, err = self.runcmd(self.cmd, "keytrust", "add",
                                       "-H", HOST, CREDS,
                                       self.user1, DUPLICATE_KEYS[1])
        self.assertCmdFail(result)
        self.assertNotIn('ATTRIBUTE_OR_VALUE_EXISTS', err)

        # adding the first file again should also fail.
        result, out, err = self.runcmd(self.cmd, "keytrust", "add",
                                       "-H", HOST, CREDS,
                                       self.user1, DUPLICATE_KEYS[0])
        self.assertCmdFail(result)

        # adding to a different DN is OK
        base_dn = self.samdb.domain_dn()
        result, out, err = self.runcmd(self.cmd, "keytrust", "add",
                                       "-H", HOST, CREDS,
                                       "--link-target", base_dn,
                                       self.user1, DUPLICATE_KEYS[1])
        self.assertCmdSuccess(result, out, err)

        self.assertEqual(len(self.get_links(self.user1)), 2)

    def test_add_wrong_size_keys(self):
        """You should not be able to add the same link twice."""

        self.addCleanup(self.runcmd, self.cmd, "keytrust", "delete",
                        "-H", HOST, CREDS,
                        self.user2, '--all')

        for fn in WRONG_SIZE_CERTS:
            result, out, err = self.runcmd(self.cmd, "keytrust", "add",
                                           "-H", HOST, CREDS,
                                           self.user2, fn)
            self.assertCmdFail(result)
            self.assertIn('ERROR: 2048 bit RSA key expected, not', err)

        self.assertEqual(self.get_links(self.user2), [])

        for fn in WRONG_SIZE_CERTS:
            # it will work with --force
            result, out, err = self.runcmd(self.cmd, "keytrust", "add",
                                           "-H", HOST, CREDS,
                                           '--force',
                                           self.user2, fn)

            self.assertCmdSuccess(result, out, err)

        self.assertEqual(len(self.get_links(self.user2)), 2)

    def test_add_non_rsa_keys(self):
        """You should not be able to add the same link twice."""

        self.addCleanup(self.runcmd, self.cmd, "keytrust", "delete",
                        "-H", HOST, CREDS,
                        self.user2, '--all')

        for fn in NON_RSA_CERTS:
            result, out, err = self.runcmd(self.cmd, "keytrust", "add",
                                           "-H", HOST, CREDS,
                                           self.user2, fn)
            self.assertCmdFail(result)
            self.assertIn('only RSA Public Keys are supported', err)

        self.assertEqual(self.get_links(self.user2), [])

        for fn in NON_RSA_CERTS:
            # it will NOT work with --force
            result, out, err = self.runcmd(self.cmd, "keytrust", "add",
                                           "-H", HOST, CREDS,
                                           '--force',
                                           self.user2, fn)

            self.assertCmdFail(result)
            self.assertIn('only RSA Public Keys are supported', err)

        self.assertEqual(self.get_links(self.user2), [])

    def test_add_good_cert_bad_dn(self):
        """Fails differently with --force"""
        links = self.get_links(self.user2)
        n = len(links)
        target = f"CN=an RDN that is not there,{self.samdb.domain_dn()}"
        result, out, err = self.runcmd(self.cmd, "keytrust", "add",
                                       "-H", HOST, CREDS,
                                       '--link-target', target,
                                       self.user2, GOOD_CERTS[0])
        self.assertCmdFail(result)
        self.assertIn(f"ERROR: Link target '{target}' does not exist", err)
        self.assertEqual(len(links), 0)

        result, out, err = self.runcmd(self.cmd, "keytrust", "add",
                                       "-H", HOST, CREDS,
                                       '--link-target', target,
                                       '--force',
                                       self.user2, GOOD_CERTS[1])
        self.assertCmdFail(result)
        self.assertIn("ERROR(ldb): uncaught exception", err)
        self.assertIn("LDAP_CONSTRAINT_VIOLATION", err)
        self.assertEqual(len(links), 0)

    def test_add_good_cert_bad_encoding(self):
        """If we use --encoding=pem with a DER file or vice versa, it
        should fail."""
        self.addCleanup(self.runcmd, self.cmd, "keytrust", "delete",
                        "-H", HOST, CREDS,
                        self.user2, '--all')

        result, out, err = self.runcmd(self.cmd, "keytrust", "add",
                                       "-H", HOST, CREDS,
                                       '--encoding', 'der',
                                       self.user2, GOOD_CERTS[0])
        self.assertCmdFail(result)
        self.assertIn("ERROR: could not decode public key", err)
        self.assertEqual(self.get_links(self.user2), [])

        # try to --force this one, to no avail
        result, out, err = self.runcmd(self.cmd, "keytrust", "add",
                                       "-H", HOST, CREDS,
                                       '--force',
                                       '--encoding', 'pem',
                                       self.user2, GOOD_KEYS[0])
        self.assertCmdFail(result)
        self.assertIn("ERROR: could not decode public key", err)
        self.assertEqual(self.get_links(self.user2), [])

        with self.assertRaises(SystemExit):
            # we can't catch result and output here because it fails
            # in optparse which prints straight to stderr.
            self.runcmd(self.cmd, "keytrust", "add",
                        "-H", HOST, CREDS,
                        '--encoding', 'pineapple',
                        self.user2, GOOD_CERTS[1])
        self.assertCmdFail(result)
        self.assertEqual(self.get_links(self.user2), [])

        # right encoding
        result, out, err = self.runcmd(self.cmd, "keytrust", "add",
                                       "-H", HOST, CREDS,
                                       '--encoding', 'pem',
                                       self.user2, GOOD_CERTS[1])
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(len(self.get_links(self.user2)), 1)

        # 'auto' encoding
        result, out, err = self.runcmd(self.cmd, "keytrust", "add",
                                       "-H", HOST, CREDS,
                                       '--encoding', 'auto',
                                       self.user2, GOOD_CERTS[0])
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(len(self.get_links(self.user2)), 2)


class SambaToolComputerKeyTrustTest(SambaToolUserKeyTrustTest):
    cmd = "computer"
    user1 = 'ADDC'
    user2 = 'ADDC'
    ou_name = 'OU=Domain Controllers'

    def get_links(self, username):
        if username[-1] != '$':
            username += '$'
        result = self.samdb.search(expression=f'sAMAccountName={username}',
                                   attrs=['msDS-KeyCredentialLink'])
        self.assertEqual(len(result), 1)
        links = result[0].get('msDS-KeyCredentialLink', [])
        return [kcl.KeyCredentialLinkDn(self.samdb, v) for v in links]



# Create schema.ldif from Github markdown
#
# Each LDF section in the markdown file then gets written to a corresponding
# .LDF output file.
#
# Copyright (C) Andrew Bartlett 2017
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Generate LDIF from Github documentation."""

import re
import os
import markdown
import xml.etree.ElementTree as ET


def innertext(tag):
    return (tag.text or '') + \
            ''.join(innertext(e) for e in tag) + \
            (tag.tail or '')


def read_ms_markdown(in_file, out_folder):
    """Read Github documentation-derived schema files."""

    with open(in_file) as update_file:
        # Remove any comments from the raw LDF files
        html = markdown.markdown(re.sub(r'(?m)^# .*\n?', '', update_file.read()),
                                 output_format='xhtml')

    tree = ET.fromstring('<root>' + html + '</root>')

    ldf = None
    try:
        for node in tree:
            if node.tag == 'h3':
                if ldf is not None:
                    ldf.close()

                out_path = os.path.join(out_folder, innertext(node).strip())
                ldf = open(out_path, 'w')
            elif node.tag == 'h2':
                if ldf is not None:
                    ldf.close()
                    ldf = None
            elif node.tag == 'p' and ldf is not None:
                ldf.write(innertext(node).replace('```', '') + '\n')
    finally:
        if ldf is not None:
            ldf.close()


if __name__ == '__main__':
    import sys

    out_folder = ''

    if len(sys.argv) == 0:
        print("Usage: %s <Schema-Update.md> [<output folder>]" % (sys.argv[0]), file=sys.stderr)
        sys.exit(1)

    in_file = sys.argv[1]
    if len(sys.argv) > 2:
        out_folder = sys.argv[2]

    read_ms_markdown(in_file, out_folder)



# -*- coding: utf-8 -*-
# Unix SMB/CIFS implementation. Tests for netbios py module
# Copyright (C) Noel Power <noel.power@suse.com> 2018
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import samba
import os
from samba import netbios


class NetBiosTests(samba.tests.TestCase):
    def setUp(self):
        super().setUp()
        self.n = netbios.Node()
        self.ifc = os.environ["SERVER_IP"]
        self.dc = os.environ["DC_NETBIOSNAME"]

    def test_query_name(self):
        (reply_from, names, addresses) = self.n.query_name(self.dc, self.ifc, timeout=4)
        assert reply_from == self.ifc
        assert names[0] == self.dc
        assert addresses[0] == self.ifc

    def test_name_status(self):
        (reply_from, name, name_list) = self.n.name_status(self.dc, self.ifc, timeout=4)
        assert reply_from == self.ifc
        assert name[0] == self.dc
        assert len(name_list) > 0

    def test_register_name(self):
        address = '127.0.0.3'
        (reply_from, name, reply_address, code) = self.n.register_name((self.dc, 0x20), address, self.ifc, multi_homed=False, timeout=4)
        assert reply_from == self.ifc
        assert name[0] == self.dc
        assert reply_address == address
        assert code == 6

    def disabled_test_refresh(self):
        # can't get the below test to work, disabling
        address = '127.0.0.3'
        res = self.n.refresh_name((self.dc, 0x20), address, self.ifc, timeout=10)


class ValidNetbiosNameTests(samba.tests.TestCase):

    def test_valid(self):
        self.assertTrue(samba.valid_netbios_name("FOO"))

    def test_too_long(self):
        self.assertFalse(samba.valid_netbios_name("FOO" * 10))

    def test_invalid_characters(self):
        self.assertFalse(samba.valid_netbios_name("*BLA"))



# domain management - domain info
#
# Copyright Matthias Dieter Wallnoefer 2009
# Copyright Andrew Kroeger 2009
# Copyright Jelmer Vernooij 2007-2012
# Copyright Giampaolo Lauria 2011
# Copyright Matthieu Patou <mat@matws.net> 2011
# Copyright Andrew Bartlett 2008-2015
# Copyright Stefan Metzmacher 2012
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import samba.getopt as options
from samba.netcmd import Command, CommandError
from samba.netcmd.common import netcmd_get_domain_infos_via_cldap


class cmd_domain_info(Command):
    """Print basic info about a domain and the DC passed as parameter."""

    synopsis = "%prog <ip_address> [options]"

    takes_options = [
    ]

    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "credopts": options.CredentialsOptions,
        "versionopts": options.VersionOptions,
    }

    takes_args = ["address"]

    def run(self, address, credopts=None, sambaopts=None, versionopts=None):
        lp = sambaopts.get_loadparm()
        try:
            res = netcmd_get_domain_infos_via_cldap(lp, None, address)
        except RuntimeError:
            raise CommandError("Invalid IP address '" + address + "'!")
        self.outf.write("Forest           : %s\n" % res.forest)
        self.outf.write("Domain           : %s\n" % res.dns_domain)
        self.outf.write("Netbios domain   : %s\n" % res.domain_name)
        self.outf.write("DC name          : %s\n" % res.pdc_dns_name)
        self.outf.write("DC netbios name  : %s\n" % res.pdc_name)
        self.outf.write("Server site      : %s\n" % res.server_site)
        self.outf.write("Client site      : %s\n" % res.client_site)



#!/usr/bin/env python3
# Unix SMB/CIFS implementation.
# Copyright (C) Stefan Metzmacher 2020
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import os

sys.path.insert(0, "bin/python")
os.environ["PYTHONUNBUFFERED"] = "1"

import functools
import time

from samba import dsdb, ntstatus
from samba.dcerpc import krb5pac, lsa, security

from samba.tests import env_get_var_value
from samba.tests.krb5.kcrypto import Cksumtype, Enctype
from samba.tests.krb5.kdc_base_test import KDCBaseTest
from samba.tests.krb5.raw_testcase import (
    RawKerberosTest,
    RodcPacEncryptionKey,
    ZeroedChecksumKey
)
from samba.tests.krb5.rfc4120_constants import (
    AD_IF_RELEVANT,
    AES256_CTS_HMAC_SHA1_96,
    ARCFOUR_HMAC_MD5,
    KDC_ERR_BADMATCH,
    KDC_ERR_BADOPTION,
    KDC_ERR_BAD_INTEGRITY,
    KDC_ERR_GENERIC,
    KDC_ERR_INAPP_CKSUM,
    KDC_ERR_MODIFIED,
    KDC_ERR_SUMTYPE_NOSUPP,
    KDC_ERR_TGT_REVOKED,
    KU_AS_REP_ENC_PART,
    KU_PA_ENC_TIMESTAMP,
    KU_TGS_REP_ENC_PART_SUB_KEY,
    KU_TGS_REQ_AUTH_DAT_SESSION,
    KU_TGS_REQ_AUTH_DAT_SUBKEY,
    NT_PRINCIPAL,
)
import samba.tests.krb5.rfc4120_pyasn1 as krb5_asn1

SidType = RawKerberosTest.SidType

global_asn1_print = False
global_hexdump = False


class S4UKerberosTests(KDCBaseTest):

    default_attrs = security.SE_GROUP_DEFAULT_FLAGS

    def setUp(self):
        super().setUp()
        self.do_asn1_print = global_asn1_print
        self.do_hexdump = global_hexdump

    def _test_s4u2self(self, pa_s4u2self_ctype=None):
        service_creds = self.get_service_creds()
        service = service_creds.get_username()
        realm = service_creds.get_realm()

        cname = self.PrincipalName_create(name_type=1, names=[service])
        sname = self.PrincipalName_create(name_type=2, names=["krbtgt", realm])

        till = self.get_KerberosTime(offset=36000)

        kdc_options = krb5_asn1.KDCOptions('forwardable')
        padata = None

        etypes = (18, 17, 23)

        req = self.AS_REQ_create(padata=padata,
                                 kdc_options=str(kdc_options),
                                 cname=cname,
                                 realm=realm,
                                 sname=sname,
                                 from_time=None,
                                 till_time=till,
                                 renew_time=None,
                                 nonce=0x7fffffff,
                                 etypes=etypes,
                                 addresses=None,
                                 additional_tickets=None)
        rep = self.send_recv_transaction(req)
        self.assertIsNotNone(rep)

        self.assertEqual(rep['msg-type'], 30)
        self.assertEqual(rep['error-code'], 25)
        rep_padata = self.der_decode(
            rep['e-data'], asn1Spec=krb5_asn1.METHOD_DATA())

        for pa in rep_padata:
            if pa['padata-type'] == 19:
                etype_info2 = pa['padata-value']
                break

        etype_info2 = self.der_decode(
            etype_info2, asn1Spec=krb5_asn1.ETYPE_INFO2())

        key = self.PasswordKey_from_etype_info2(service_creds, etype_info2[0])

        (patime, pausec) = self.get_KerberosTimeWithUsec()
        pa_ts = self.PA_ENC_TS_ENC_create(patime, pausec)
        pa_ts = self.der_encode(pa_ts, asn1Spec=krb5_asn1.PA_ENC_TS_ENC())

        pa_ts = self.EncryptedData_create(key, KU_PA_ENC_TIMESTAMP, pa_ts)
        pa_ts = self.der_encode(pa_ts, asn1Spec=krb5_asn1.EncryptedData())

        pa_ts = self.PA_DATA_create(2, pa_ts)

        kdc_options = krb5_asn1.KDCOptions('forwardable')
        padata = [pa_ts]

        req = self.AS_REQ_create(padata=padata,
                                 kdc_options=str(kdc_options),
                                 cname=cname,
                                 realm=realm,
                                 sname=sname,
                                 from_time=None,
                                 till_time=till,
                                 renew_time=None,
                                 nonce=0x7fffffff,
                                 etypes=etypes,
                                 addresses=None,
                                 additional_tickets=None)
        rep = self.send_recv_transaction(req)
        self.assertIsNotNone(rep)

        msg_type = rep['msg-type']
        self.assertEqual(msg_type, 11)

        enc_part2 = key.decrypt(KU_AS_REP_ENC_PART, rep['enc-part']['cipher'])
        # MIT KDC encodes both EncASRepPart and EncTGSRepPart with
        # application tag 26
        try:
            enc_part2 = self.der_decode(
                enc_part2, asn1Spec=krb5_asn1.EncASRepPart())
        except Exception:
            enc_part2 = self.der_decode(
                enc_part2, asn1Spec=krb5_asn1.EncTGSRepPart())

        # S4U2Self Request
        sname = cname

        for_user_name = env_get_var_value('FOR_USER')
        uname = self.PrincipalName_create(name_type=1, names=[for_user_name])

        kdc_options = krb5_asn1.KDCOptions('forwardable')
        till = self.get_KerberosTime(offset=36000)
        ticket = rep['ticket']
        ticket_session_key = self.EncryptionKey_import(enc_part2['key'])
        pa_s4u = self.PA_S4U2Self_create(name=uname, realm=realm,
                                         tgt_session_key=ticket_session_key,
                                         ctype=pa_s4u2self_ctype)
        padata = [pa_s4u]

        subkey = self.RandomKey(ticket_session_key.etype)

        (ctime, cusec) = self.get_KerberosTimeWithUsec()

        req = self.TGS_REQ_create(padata=padata,
                                  cusec=cusec,
                                  ctime=ctime,
                                  ticket=ticket,
                                  kdc_options=str(kdc_options),
                                  cname=cname,
                                  realm=realm,
                                  sname=sname,
                                  from_time=None,
                                  till_time=till,
                                  renew_time=None,
                                  nonce=0x7ffffffe,
                                  etypes=etypes,
                                  addresses=None,
                                  EncAuthorizationData=None,
                                  EncAuthorizationData_key=None,
                                  additional_tickets=None,
                                  ticket_session_key=ticket_session_key,
                                  authenticator_subkey=subkey)
        rep = self.send_recv_transaction(req)
        self.assertIsNotNone(rep)

        msg_type = rep['msg-type']
        if msg_type == 13:
            enc_part2 = subkey.decrypt(
                KU_TGS_REP_ENC_PART_SUB_KEY, rep['enc-part']['cipher'])
            enc_part2 = self.der_decode(
                enc_part2, asn1Spec=krb5_asn1.EncTGSRepPart())

        return msg_type

    # Using the checksum type from the tgt_session_key happens to work
    # everywhere
    def test_s4u2self(self):
        msg_type = self._test_s4u2self()
        self.assertEqual(msg_type, 13)

    # Per spec, the checksum of PA-FOR-USER is HMAC_MD5, see [MS-SFU] 2.2.1
    def test_s4u2self_hmac_md5_checksum(self):
        msg_type = self._test_s4u2self(pa_s4u2self_ctype=Cksumtype.HMAC_MD5)
        self.assertEqual(msg_type, 13)

    def test_s4u2self_md5_unkeyed_checksum(self):
        msg_type = self._test_s4u2self(pa_s4u2self_ctype=Cksumtype.MD5)
        self.assertEqual(msg_type, 30)

    def test_s4u2self_sha1_unkeyed_checksum(self):
        msg_type = self._test_s4u2self(pa_s4u2self_ctype=Cksumtype.SHA1)
        self.assertEqual(msg_type, 30)

    def test_s4u2self_crc32_unkeyed_checksum(self):
        msg_type = self._test_s4u2self(pa_s4u2self_ctype=Cksumtype.CRC32)
        self.assertEqual(msg_type, 30)

    def _run_s4u2self_test(self, kdc_dict):
        client_opts = kdc_dict.pop('client_opts', None)
        client_creds = self.get_cached_creds(
            account_type=self.AccountType.USER,
            opts=client_opts)

        service_opts = kdc_dict.pop('service_opts', None)
        service_creds = self.get_cached_creds(
            account_type=self.AccountType.COMPUTER,
            opts=service_opts)

        service_tgt = self.get_tgt(service_creds)
        modify_service_tgt_fn = kdc_dict.pop('modify_service_tgt_fn', None)
        if modify_service_tgt_fn is not None:
            service_tgt = modify_service_tgt_fn(service_tgt)

        client_name = client_creds.get_username()
        client_cname = self.PrincipalName_create(name_type=NT_PRINCIPAL,
                                                 names=[client_name])

        service_name = kdc_dict.pop('service_name', None)
        if service_name is None:
            service_name = service_creds.get_username()[:-1]
        service_sname = self.PrincipalName_create(name_type=NT_PRINCIPAL,
                                                  names=['host', service_name])

        realm = client_creds.get_realm()

        expected_flags = kdc_dict.pop('expected_flags', None)
        if expected_flags is not None:
            expected_flags = krb5_asn1.TicketFlags(expected_flags)

        unexpected_flags = kdc_dict.pop('unexpected_flags', None)
        if unexpected_flags is not None:
            unexpected_flags = krb5_asn1.TicketFlags(unexpected_flags)

        expected_error_mode = kdc_dict.pop('expected_error_mode', 0)
        expect_status = kdc_dict.pop('expect_status', None)
        expected_status = kdc_dict.pop('expected_status', None)
        if expected_error_mode:
            check_error_fn = self.generic_check_kdc_error
            check_rep_fn = None
        else:
            check_error_fn = None
            check_rep_fn = self.generic_check_kdc_rep

            self.assertIsNone(expect_status)
            self.assertIsNone(expected_status)

        kdc_options = kdc_dict.pop('kdc_options', '0')
        kdc_options = krb5_asn1.KDCOptions(kdc_options)

        service_decryption_key = self.TicketDecryptionKey_from_creds(
            service_creds)

        authenticator_subkey = self.RandomKey(Enctype.AES256)

        etypes = kdc_dict.pop('etypes', (AES256_CTS_HMAC_SHA1_96,
                                         ARCFOUR_HMAC_MD5))

        expect_edata = kdc_dict.pop('expect_edata', None)
        expected_groups = kdc_dict.pop('expected_groups', None)
        unexpected_groups = kdc_dict.pop('unexpected_groups', None)

        def generate_s4u2self_padata(_kdc_exchange_dict,
                                     _callback_dict,
                                     req_body):
            pa_s4u = self.PA_S4U2Self_create(
                name=client_cname,
                realm=realm,
                tgt_session_key=service_tgt.session_key,
                ctype=None)

            return [pa_s4u], req_body

        kdc_exchange_dict = self.tgs_exchange_dict(
            expected_crealm=realm,
            expected_cname=client_cname,
            expected_srealm=realm,
            expected_sname=service_sname,
            expected_account_name=client_name,
            expected_groups=expected_groups,
            unexpected_groups=unexpected_groups,
            expected_sid=client_creds.get_sid(),
            expected_flags=expected_flags,
            unexpected_flags=unexpected_flags,
            ticket_decryption_key=service_decryption_key,
            expect_ticket_checksum=True,
            generate_padata_fn=generate_s4u2self_padata,
            check_error_fn=check_error_fn,
            check_rep_fn=check_rep_fn,
            check_kdc_private_fn=self.generic_check_kdc_private,
            expected_error_mode=expected_error_mode,
            expect_status=expect_status,
            expected_status=expected_status,
            tgt=service_tgt,
            authenticator_subkey=authenticator_subkey,
            kdc_options=str(kdc_options),
            expect_edata=expect_edata)

        self._generic_kdc_exchange(kdc_exchange_dict,
                                   cname=None,
                                   realm=realm,
                                   sname=service_sname,
                                   etypes=etypes)

        if not expected_error_mode:
            # Check that the ticket contains a PAC.
            ticket = kdc_exchange_dict['rep_ticket_creds']

            pac = self.get_ticket_pac(ticket)
            self.assertIsNotNone(pac)

        # Ensure we used all the parameters given to us.
        self.assertEqual({}, kdc_dict)

    # Test performing an S4U2Self operation with a forwardable ticket. The
    # resulting ticket should have the 'forwardable' flag set.
    def test_s4u2self_forwardable(self):
        self._run_s4u2self_test(
            {
                'client_opts': {
                    'not_delegated': False
                },
                'kdc_options': 'forwardable',
                'modify_service_tgt_fn': functools.partial(
                    self.set_ticket_forwardable, flag=True),
                'expected_flags': 'forwardable'
            })

    # Test performing an S4U2Self operation with a forwardable ticket that does
    # not contain a PAC. The request should fail.
    def test_s4u2self_no_pac(self):
        def forwardable_no_pac(ticket):
            ticket = self.set_ticket_forwardable(ticket, flag=True)
            return self.remove_ticket_pac(ticket)

        self._run_s4u2self_test(
            {
                'expected_error_mode': KDC_ERR_TGT_REVOKED,
                'client_opts': {
                    'not_delegated': False
                },
                'kdc_options': 'forwardable',
                'modify_service_tgt_fn': forwardable_no_pac,
                'expected_flags': 'forwardable',
                'expect_edata': False
            })

    # Test performing an S4U2Self operation without requesting a forwardable
    # ticket. The resulting ticket should not have the 'forwardable' flag set.
    def test_s4u2self_without_forwardable(self):
        self._run_s4u2self_test(
            {
                'client_opts': {
                    'not_delegated': False
                },
                'modify_service_tgt_fn': functools.partial(
                    self.set_ticket_forwardable, flag=True),
                'unexpected_flags': 'forwardable'
            })

    # Do an S4U2Self with a non-forwardable TGT. The 'forwardable' flag should
    # not be set on the ticket.
    def test_s4u2self_not_forwardable(self):
        self._run_s4u2self_test(
            {
                'client_opts': {
                    'not_delegated': False
                },
                'kdc_options': 'forwardable',
                'modify_service_tgt_fn': functools.partial(
                    self.set_ticket_forwardable, flag=False),
                'unexpected_flags': 'forwardable'
            })

    # Do an S4U2Self with the not_delegated flag set on the client. The
    # 'forwardable' flag should not be set on the ticket.
    def test_s4u2self_client_not_delegated(self):
        self._run_s4u2self_test(
            {
                'client_opts': {
                    'not_delegated': True
                },
                'kdc_options': 'forwardable',
                'modify_service_tgt_fn': functools.partial(
                    self.set_ticket_forwardable, flag=True),
                'unexpected_flags': 'forwardable'
            })

    # Do an S4U2Self with a service not trusted to authenticate for delegation,
    # but having an empty msDS-AllowedToDelegateTo attribute. The 'forwardable'
    # flag should be set on the ticket.
    def test_s4u2self_not_trusted_empty_allowed(self):
        self._run_s4u2self_test(
            {
                'client_opts': {
                    'not_delegated': False
                },
                'service_opts': {
                    'trusted_to_auth_for_delegation': False,
                    'delegation_to_spn': ()
                },
                'kdc_options': 'forwardable',
                'modify_service_tgt_fn': functools.partial(
                    self.set_ticket_forwardable, flag=True),
                'expected_flags': 'forwardable'
            })

    # Do an S4U2Self with a service not trusted to authenticate for delegation
    # and having a non-empty msDS-AllowedToDelegateTo attribute. The
    # 'forwardable' flag should not be set on the ticket.
    def test_s4u2self_not_trusted_nonempty_allowed(self):
        self._run_s4u2self_test(
            {
                'client_opts': {
                    'not_delegated': False
                },
                'service_opts': {
                    'trusted_to_auth_for_delegation': False,
                    'delegation_to_spn': ('test',)
                },
                'kdc_options': 'forwardable',
                'modify_service_tgt_fn': functools.partial(
                    self.set_ticket_forwardable, flag=True),
                'unexpected_flags': 'forwardable'
            })

    # Do an S4U2Self with a service trusted to authenticate for delegation and
    # having an empty msDS-AllowedToDelegateTo attribute. The 'forwardable'
    # flag should be set on the ticket.
    def test_s4u2self_trusted_empty_allowed(self):
        self._run_s4u2self_test(
            {
                'client_opts': {
                    'not_delegated': False
                },
                'service_opts': {
                    'trusted_to_auth_for_delegation': True,
                    'delegation_to_spn': ()
                },
                'kdc_options': 'forwardable',
                'modify_service_tgt_fn': functools.partial(
                    self.set_ticket_forwardable, flag=True),
                'expected_flags': 'forwardable'
            })

    # Do an S4U2Self with a service trusted to authenticate for delegation and
    # having a non-empty msDS-AllowedToDelegateTo attribute. The 'forwardable'
    # flag should be set on the ticket.
    def test_s4u2self_trusted_nonempty_allowed(self):
        self._run_s4u2self_test(
            {
                'client_opts': {
                    'not_delegated': False
                },
                'service_opts': {
                    'trusted_to_auth_for_delegation': True,
                    'delegation_to_spn': ('test',)
                },
                'kdc_options': 'forwardable',
                'modify_service_tgt_fn': functools.partial(
                    self.set_ticket_forwardable, flag=True),
                'expected_flags': 'forwardable'
            })

    # Do an S4U2Self with the sname in the request different to that of the
    # service. We expect an error.
    def test_s4u2self_wrong_sname(self):
        other_creds = self.get_cached_creds(
            account_type=self.AccountType.COMPUTER,
            opts={
                'trusted_to_auth_for_delegation': True,
                'id': 0
            })
        other_sname = other_creds.get_username()[:-1]

        self._run_s4u2self_test(
            {
                'expected_error_mode': KDC_ERR_BADMATCH,
                'expect_edata': False,
                'client_opts': {
                    'not_delegated': False
                },
                'service_opts': {
                    'trusted_to_auth_for_delegation': True
                },
                'service_name': other_sname,
                'kdc_options': 'forwardable',
                'modify_service_tgt_fn': functools.partial(
                    self.set_ticket_forwardable, flag=True)
            })

    # Do an S4U2Self where the service does not require authorization data. The
    # resulting ticket should still contain a PAC.
    def test_s4u2self_no_auth_data_required(self):
        self._run_s4u2self_test(
            {
                'client_opts': {
                    'not_delegated': False
                },
                'service_opts': {
                    'trusted_to_auth_for_delegation': True,
                    'no_auth_data_required': True
                },
                'kdc_options': 'forwardable',
                'modify_service_tgt_fn': functools.partial(
                    self.set_ticket_forwardable, flag=True),
                'expected_flags': 'forwardable'
            })

    # Do an S4U2Self and check that the service asserted identity is part of
    # the sids.
    def test_s4u2self_asserted_identity(self):
        self._run_s4u2self_test(
            {
                'client_opts': {
                    'not_delegated': False
                },
                'expected_groups': {
                    (security.SID_SERVICE_ASSERTED_IDENTITY,
                     SidType.EXTRA_SID,
                     self.default_attrs),
                    ...
                },
                'unexpected_groups': {
                    security.SID_AUTHENTICATION_AUTHORITY_ASSERTED_IDENTITY,
                },
            })

    def _run_delegation_test(self, kdc_dict):
        s4u2self = kdc_dict.pop('s4u2self', False)

        authtime_delay = kdc_dict.pop('authtime_delay', 0)

        client_opts = kdc_dict.pop('client_opts', None)
        client_creds = self.get_cached_creds(
            account_type=self.AccountType.USER,
            opts=client_opts)

        sid = client_creds.get_sid()

        service1_opts = kdc_dict.pop('service1_opts', {})
        service2_opts = kdc_dict.pop('service2_opts', {})

        allow_delegation = kdc_dict.pop('allow_delegation', False)
        allow_rbcd = kdc_dict.pop('allow_rbcd', False)
        self.assertFalse(allow_delegation and allow_rbcd)

        if allow_rbcd:
            service1_creds = self.get_cached_creds(
                account_type=self.AccountType.COMPUTER,
                opts=service1_opts)

            self.assertNotIn('delegation_from_dn', service2_opts)
            service2_opts['delegation_from_dn'] = str(service1_creds.get_dn())

            service2_creds = self.get_cached_creds(
                account_type=self.AccountType.COMPUTER,
                opts=service2_opts)
        else:
            service2_creds = self.get_cached_creds(
                account_type=self.AccountType.COMPUTER,
                opts=service2_opts)

            if allow_delegation:
                self.assertNotIn('delegation_to_spn', service1_opts)
                service1_opts['delegation_to_spn'] = service2_creds.get_spn()

            service1_creds = self.get_cached_creds(
                account_type=self.AccountType.COMPUTER,
                opts=service1_opts)

        service1_tgt = self.get_tgt(service1_creds)
        self.assertElementPresent(service1_tgt.ticket_private, 'authtime')
        service1_tgt_authtime = self.getElementValue(service1_tgt.ticket_private, 'authtime')

        client_username = client_creds.get_username()
        client_realm = client_creds.get_realm()
        client_cname = self.PrincipalName_create(name_type=NT_PRINCIPAL,
                                                 names=[client_username])

        service1_name = service1_creds.get_username()[:-1]
        service1_realm = service1_creds.get_realm()
        service1_service = 'host'
        service1_sname = self.PrincipalName_create(
            name_type=NT_PRINCIPAL, names=[service1_service,
                                           service1_name])
        service1_decryption_key = self.TicketDecryptionKey_from_creds(
            service1_creds)

        expect_pac = kdc_dict.pop('expect_pac', True)

        expected_groups = kdc_dict.pop('expected_groups', None)
        unexpected_groups = kdc_dict.pop('unexpected_groups', None)

        client_tkt_options = kdc_dict.pop('client_tkt_options', 'forwardable')
        expected_flags = krb5_asn1.TicketFlags(client_tkt_options)

        etypes = kdc_dict.pop('etypes', (AES256_CTS_HMAC_SHA1_96,
                                         ARCFOUR_HMAC_MD5))

        if s4u2self:
            self.assertEqual(authtime_delay, 0)

            def generate_s4u2self_padata(_kdc_exchange_dict,
                                         _callback_dict,
                                         req_body):
                pa_s4u = self.PA_S4U2Self_create(
                    name=client_cname,
                    realm=client_realm,
                    tgt_session_key=service1_tgt.session_key,
                    ctype=None)

                return [pa_s4u], req_body

            s4u2self_expected_flags = krb5_asn1.TicketFlags('forwardable')
            s4u2self_unexpected_flags = krb5_asn1.TicketFlags('0')

            s4u2self_kdc_options = krb5_asn1.KDCOptions('forwardable')

            s4u2self_authenticator_subkey = self.RandomKey(Enctype.AES256)
            s4u2self_kdc_exchange_dict = self.tgs_exchange_dict(
                expected_crealm=client_realm,
                expected_cname=client_cname,
                expected_srealm=service1_realm,
                expected_sname=service1_sname,
                expected_account_name=client_username,
                expected_groups=expected_groups,
                unexpected_groups=unexpected_groups,
                expected_sid=sid,
                expected_flags=s4u2self_expected_flags,
                unexpected_flags=s4u2self_unexpected_flags,
                ticket_decryption_key=service1_decryption_key,
                generate_padata_fn=generate_s4u2self_padata,
                check_rep_fn=self.generic_check_kdc_rep,
                check_kdc_private_fn=self.generic_check_kdc_private,
                tgt=service1_tgt,
                authenticator_subkey=s4u2self_authenticator_subkey,
                kdc_options=str(s4u2self_kdc_options),
                expect_edata=False)

            self._generic_kdc_exchange(s4u2self_kdc_exchange_dict,
                                       cname=None,
                                       realm=service1_realm,
                                       sname=service1_sname,
                                       etypes=etypes)

            client_service_tkt = s4u2self_kdc_exchange_dict['rep_ticket_creds']
        else:
            if authtime_delay != 0:
                time.sleep(authtime_delay)
                fresh = True
            else:
                fresh = False

            client_tgt = self.get_tgt(client_creds,
                                      kdc_options=client_tkt_options,
                                      expected_flags=expected_flags,
                                      fresh=fresh)
            client_service_tkt = self.get_service_ticket(
                client_tgt,
                service1_creds,
                kdc_options=client_tkt_options,
                expected_flags=expected_flags,
                fresh=fresh)

        modify_client_tkt_fn = kdc_dict.pop('modify_client_tkt_fn', None)
        if modify_client_tkt_fn is not None:
            client_service_tkt = modify_client_tkt_fn(client_service_tkt)

        self.assertElementPresent(client_service_tkt.ticket_private, 'authtime')
        expected_authtime = self.getElementValue(client_service_tkt.ticket_private, 'authtime')
        if authtime_delay > 1:
            self.assertNotEqual(expected_authtime, service1_tgt_authtime)

        additional_tickets = [client_service_tkt.ticket]

        modify_service_tgt_fn = kdc_dict.pop('modify_service_tgt_fn', None)
        if modify_service_tgt_fn is not None:
            service1_tgt = modify_service_tgt_fn(service1_tgt)

        kdc_options = kdc_dict.pop('kdc_options', None)
        if kdc_options is None:
            kdc_options = str(krb5_asn1.KDCOptions('cname-in-addl-tkt'))

        service2_name = service2_creds.get_username()[:-1]
        service2_realm = service2_creds.get_realm()
        service2_service = 'host'
        service2_sname = self.PrincipalName_create(
            name_type=NT_PRINCIPAL, names=[service2_service,
                                           service2_name])
        service2_decryption_key = self.TicketDecryptionKey_from_creds(
            service2_creds)
        service2_etypes = service2_creds.tgs_supported_enctypes

        expected_error_mode = kdc_dict.pop('expected_error_mode')
        expect_status = kdc_dict.pop('expect_status', None)
        expected_status = kdc_dict.pop('expected_status', None)
        if expected_error_mode:
            check_error_fn = self.generic_check_kdc_error
            check_rep_fn = None
        else:
            check_error_fn = None
            check_rep_fn = self.generic_check_kdc_rep

            self.assertIsNone(expect_status)
            self.assertIsNone(expected_status)

        expect_edata = kdc_dict.pop('expect_edata', None)
        if expect_edata is not None:
            self.assertTrue(expected_error_mode)

        pac_options = kdc_dict.pop('pac_options', None)

        use_authenticator_subkey = kdc_dict.pop('use_authenticator_subkey', True)
        if use_authenticator_subkey:
            authenticator_subkey = self.RandomKey(Enctype.AES256)
        else:
            authenticator_subkey = None

        expected_proxy_target = service2_creds.get_spn()

        expected_transited_services = kdc_dict.pop(
            'expected_transited_services', [])

        transited_service = f'host/{service1_name}@{service1_realm}'
        expected_transited_services.append(transited_service)

        kdc_exchange_dict = self.tgs_exchange_dict(
            expected_crealm=client_realm,
            expected_cname=client_cname,
            expected_srealm=service2_realm,
            expected_sname=service2_sname,
            expected_account_name=client_username,
            expected_groups=expected_groups,
            unexpected_groups=unexpected_groups,
            expected_sid=sid,
            expected_supported_etypes=service2_etypes,
            ticket_decryption_key=service2_decryption_key,
            check_error_fn=check_error_fn,
            check_rep_fn=check_rep_fn,
            check_kdc_private_fn=self.generic_check_kdc_private,
            expected_error_mode=expected_error_mode,
            expect_status=expect_status,
            expected_status=expected_status,
            callback_dict={},
            tgt=service1_tgt,
            authenticator_subkey=authenticator_subkey,
            kdc_options=kdc_options,
            pac_options=pac_options,
            expect_edata=expect_edata,
            expected_proxy_target=expected_proxy_target,
            expected_transited_services=expected_transited_services,
            expect_pac=expect_pac)

        EncAuthorizationData = kdc_dict.pop('enc-authorization-data', None)

        if EncAuthorizationData is not None:
            if authenticator_subkey is not None:
                EncAuthorizationData_key = authenticator_subkey
                EncAuthorizationData_usage = KU_TGS_REQ_AUTH_DAT_SUBKEY
            else:
                EncAuthorizationData_key = client_service_tkt.session_key
                EncAuthorizationData_usage = KU_TGS_REQ_AUTH_DAT_SESSION
        else:
            EncAuthorizationData_key = None
            EncAuthorizationData_usage = None

        self._generic_kdc_exchange(kdc_exchange_dict,
                                   cname=None,
                                   realm=service2_realm,
                                   sname=service2_sname,
                                   etypes=etypes,
                                   additional_tickets=additional_tickets,
                                   EncAuthorizationData=EncAuthorizationData,
                                   EncAuthorizationData_key=EncAuthorizationData_key,
                                   EncAuthorizationData_usage=EncAuthorizationData_usage)

        if not expected_error_mode:
            # Check whether the ticket contains a PAC.
            ticket = kdc_exchange_dict['rep_ticket_creds']
            self.assertElementEqual(ticket.ticket_private, 'authtime', expected_authtime)
            pac = self.get_ticket_pac(ticket, expect_pac=expect_pac)
            ticket_auth_data = ticket.ticket_private.get('authorization-data')
            expected_num_ticket_auth_data = 0
            if expect_pac:
                self.assertIsNotNone(pac)
                expected_num_ticket_auth_data += 1
            else:
                self.assertIsNone(pac)

            if EncAuthorizationData is not None:
                expected_num_ticket_auth_data += len(EncAuthorizationData)

            if expected_num_ticket_auth_data == 0:
                self.assertIsNone(ticket_auth_data)
            else:
                self.assertIsNotNone(ticket_auth_data)
                self.assertEqual(len(ticket_auth_data),
                        expected_num_ticket_auth_data)

                if EncAuthorizationData is not None:
                    enc_ad_plain = self.der_encode(
                        EncAuthorizationData,
                        asn1Spec=krb5_asn1.AuthorizationData())
                    req_EncAuthorizationData = self.der_decode(
                        enc_ad_plain,
                        asn1Spec=krb5_asn1.AuthorizationData())

                    rep_EncAuthorizationData = ticket_auth_data.copy()
                    if expect_pac:
                        rep_EncAuthorizationData.pop(0)
                    self.assertEqual(rep_EncAuthorizationData, req_EncAuthorizationData)

        # Ensure we used all the parameters given to us.
        self.assertEqual({}, kdc_dict)

    def skip_unless_fl2008(self):
        samdb = self.get_samdb()
        functional_level = self.get_domain_functional_level(samdb)

        if functional_level < dsdb.DS_DOMAIN_FUNCTION_2008:
            self.skipTest('RBCD requires FL2008')

    def test_constrained_delegation(self):
        # Test constrained delegation.
        self._run_delegation_test(
            {
                'expected_error_mode': 0,
                'allow_delegation': True
            })

    def test_constrained_delegation_authtime(self):
        # Test constrained delegation.
        self._run_delegation_test(
            {
                'expected_error_mode': 0,
                'allow_delegation': True,
                'authtime_delay': 2,
            })

    def test_constrained_delegation_with_enc_auth_data_subkey(self):
        # Test constrained delegation.
        EncAuthorizationData = []
        relevant_elems = []
        auth_data777 = self.AuthorizationData_create(777, b'AuthorizationData777')
        relevant_elems.append(auth_data777)
        auth_data999 = self.AuthorizationData_create(999, b'AuthorizationData999')
        relevant_elems.append(auth_data999)
        ad_relevant = self.der_encode(relevant_elems, asn1Spec=krb5_asn1.AD_IF_RELEVANT())
        ad_data = self.AuthorizationData_create(AD_IF_RELEVANT, ad_relevant)
        EncAuthorizationData.append(ad_data)
        self._run_delegation_test(
            {
                'expected_error_mode': 0,
                'allow_delegation': True,
                'enc-authorization-data': EncAuthorizationData,
            })

    def test_constrained_delegation_with_enc_auth_data_no_subkey(self):
        # Test constrained delegation.
        EncAuthorizationData = []
        relevant_elems = []
        auth_data777 = self.AuthorizationData_create(777, b'AuthorizationData777')
        relevant_elems.append(auth_data777)
        auth_data999 = self.AuthorizationData_create(999, b'AuthorizationData999')
        relevant_elems.append(auth_data999)
        ad_relevant = self.der_encode(relevant_elems, asn1Spec=krb5_asn1.AD_IF_RELEVANT())
        ad_data = self.AuthorizationData_create(AD_IF_RELEVANT, ad_relevant)
        EncAuthorizationData.append(ad_data)
        self._run_delegation_test(
            {
                'expected_error_mode': 0,
                'allow_delegation': True,
                'enc-authorization-data': EncAuthorizationData,
                'use_authenticator_subkey': False,
            })

    def test_constrained_delegation_authentication_asserted_identity(self):
        # Test constrained delegation and check asserted identity is the
        # authentication authority. Note that we should always find this
        # SID for all the requests. Just S4U2Self will have a different SID.
        self._run_delegation_test(
            {
                'expected_error_mode': 0,
                'allow_delegation': True,
                'expected_groups': {
                    (security.SID_AUTHENTICATION_AUTHORITY_ASSERTED_IDENTITY,
                     SidType.EXTRA_SID,
                     self.default_attrs),
                    ...
                },
                'unexpected_groups': {
                    security.SID_SERVICE_ASSERTED_IDENTITY,
                },
            })

    def test_constrained_delegation_service_asserted_identity(self):
        # Test constrained delegation and check asserted identity is the
        # service sid is there. This is a S4U2Proxy + S4U2Self test.
        self._run_delegation_test(
            {
                'expected_error_mode': 0,
                'allow_delegation': True,
                's4u2self': True,
                'service1_opts': {
                    'trusted_to_auth_for_delegation': True,
                },
                'expected_groups': {
                    (security.SID_SERVICE_ASSERTED_IDENTITY,
                     SidType.EXTRA_SID,
                     self.default_attrs),
                    ...
                },
                'unexpected_groups': {
                    security.SID_AUTHENTICATION_AUTHORITY_ASSERTED_IDENTITY,
                },
            })

    def test_constrained_delegation_no_auth_data_required(self):
        # Test constrained delegation.
        self._run_delegation_test(
            {
                'expected_error_mode': 0,
                'allow_delegation': True,
                'service2_opts': {
                    'no_auth_data_required': True
                },
                'expect_pac': False
            })

    def test_constrained_delegation_existing_delegation_info(self):
        # Test constrained delegation with an existing S4U_DELEGATION_INFO
        # structure in the PAC.

        services = ['service1', 'service2', 'service3']

        self._run_delegation_test(
            {
                'expected_error_mode': 0,
                'allow_delegation': True,
                'modify_client_tkt_fn': functools.partial(
                    self.add_delegation_info, services=services),
                'expected_transited_services': services
            })

    def test_constrained_delegation_not_allowed(self):
        # Test constrained delegation when the delegating service does not
        # allow it.
        self._run_delegation_test(
            {
                'expected_error_mode': KDC_ERR_BADOPTION,
                # We aren’t particular about whether or not we get an NTSTATUS.
                'expect_status': None,
                'expected_status': ntstatus.NT_STATUS_NOT_SUPPORTED,
                'allow_delegation': False
            })

    def test_constrained_delegation_no_client_pac(self):
        # Test constrained delegation when the client service ticket does not
        # contain a PAC.
        self._run_delegation_test(
            {
                'expected_error_mode': (KDC_ERR_MODIFIED,
                                        KDC_ERR_TGT_REVOKED),
                'allow_delegation': True,
                'modify_client_tkt_fn': self.remove_ticket_pac,
                'expect_edata': False
            })

    def test_constrained_delegation_no_service_pac(self):
        # Test constrained delegation when the service TGT does not contain a
        # PAC.
        self._run_delegation_test(
            {
                'expected_error_mode': KDC_ERR_TGT_REVOKED,
                'allow_delegation': True,
                'modify_service_tgt_fn': self.remove_ticket_pac,
                'expect_edata': False
            })

    def test_constrained_delegation_no_client_pac_no_auth_data_required(self):
        # Test constrained delegation when the client service ticket does not
        # contain a PAC.
        self._run_delegation_test(
            {
                'expected_error_mode': (KDC_ERR_MODIFIED,
                                        KDC_ERR_BADOPTION,
                                        KDC_ERR_TGT_REVOKED),
                'allow_delegation': True,
                'modify_client_tkt_fn': self.remove_ticket_pac,
                'expect_edata': False,
                'service2_opts': {
                    'no_auth_data_required': True
                }
            })

    def test_constrained_delegation_no_service_pac_no_auth_data_required(self):
        # Test constrained delegation when the service TGT does not contain a
        # PAC.
        self._run_delegation_test(
            {
                'expected_error_mode': KDC_ERR_TGT_REVOKED,
                'allow_delegation': True,
                'modify_service_tgt_fn': self.remove_ticket_pac,
                'service2_opts': {
                    'no_auth_data_required': True
                },
                'expect_pac': False,
                'expect_edata': False
            })

    def test_constrained_delegation_non_forwardable(self):
        # Test constrained delegation with a non-forwardable ticket.
        self._run_delegation_test(
            {
                'expected_error_mode': KDC_ERR_BADOPTION,
                # We aren’t particular about whether or not we get an NTSTATUS.
                'expect_status': None,
                'expected_status': ntstatus.NT_STATUS_ACCOUNT_RESTRICTION,
                'allow_delegation': True,
                'modify_client_tkt_fn': functools.partial(
                    self.set_ticket_forwardable, flag=False)
            })

    def test_constrained_delegation_pac_options_rbcd(self):
        # Test constrained delegation, but with the RBCD bit set in the PAC
        # options.
        self._run_delegation_test(
            {
                'expected_error_mode': 0,
                'pac_options': '0001',  # supports RBCD
                'allow_delegation': True
            })

    def test_rbcd(self):
        # Test resource-based constrained delegation.
        self.skip_unless_fl2008()

        self._run_delegation_test(
            {
                'expected_error_mode': 0,
                'allow_rbcd': True,
                'pac_options': '0001',  # supports RBCD
            })

    def test_rbcd_no_auth_data_required(self):
        self.skip_unless_fl2008()

        self._run_delegation_test(
            {
                'expected_error_mode': 0,
                'allow_rbcd': True,
                'pac_options': '0001',  # supports RBCD
                'service2_opts': {
                    'no_auth_data_required': True
                },
                'expect_pac': False
            })

    def test_rbcd_existing_delegation_info(self):
        self.skip_unless_fl2008()

        # Test constrained delegation with an existing S4U_DELEGATION_INFO
        # structure in the PAC.

        services = ['service1', 'service2', 'service3']

        self._run_delegation_test(
            {
                'expected_error_mode': 0,
                'allow_rbcd': True,
                'pac_options': '0001',  # supports RBCD
                'modify_client_tkt_fn': functools.partial(
                    self.add_delegation_info, services=services),
                'expected_transited_services': services
            })

    def test_rbcd_not_allowed(self):
        # Test resource-based constrained delegation when the target service
        # does not allow it.
        self._run_delegation_test(
            {
                'expected_error_mode': KDC_ERR_BADOPTION,
                # We aren’t particular about whether or not we get an NTSTATUS.
                'expect_status': None,
                'expected_status': ntstatus.NT_STATUS_NOT_FOUND,
                'allow_rbcd': False,
                'pac_options': '0001'  # supports RBCD
            })

    def test_rbcd_no_client_pac_a(self):
        self.skip_unless_fl2008()

        # Test constrained delegation when the client service ticket does not
        # contain a PAC, and an empty msDS-AllowedToDelegateTo attribute.
        self._run_delegation_test(
            {
                'expected_error_mode': (KDC_ERR_MODIFIED,
                                        KDC_ERR_TGT_REVOKED),
                # We aren’t particular about whether or not we get an NTSTATUS.
                'expect_status': None,
                'expected_status': ntstatus.NT_STATUS_NOT_SUPPORTED,
                'allow_rbcd': True,
                'pac_options': '0001',  # supports RBCD
                'modify_client_tkt_fn': self.remove_ticket_pac
            })

    def test_rbcd_no_client_pac_b(self):
        self.skip_unless_fl2008()

        # Test constrained delegation when the client service ticket does not
        # contain a PAC, and a non-empty msDS-AllowedToDelegateTo attribute.
        self._run_delegation_test(
            {
                'expected_error_mode': (KDC_ERR_MODIFIED,
                                        KDC_ERR_TGT_REVOKED),
                # We aren’t particular about whether or not we get an NTSTATUS.
                'expect_status': None,
                'expected_status': ntstatus.NT_STATUS_NO_MATCH,
                'allow_rbcd': True,
                'pac_options': '0001',  # supports RBCD
                'modify_client_tkt_fn': self.remove_ticket_pac,
                'service1_opts': {
                    'delegation_to_spn': ('host/test')
                }
            })

    def test_rbcd_no_service_pac(self):
        self.skip_unless_fl2008()

        # Test constrained delegation when the service TGT does not contain a
        # PAC.
        self._run_delegation_test(
            {
                'expected_error_mode': KDC_ERR_TGT_REVOKED,
                'allow_rbcd': True,
                'pac_options': '0001',  # supports RBCD
                'modify_service_tgt_fn': self.remove_ticket_pac,
                'expect_edata': False
            })

    def test_rbcd_no_client_pac_no_auth_data_required_a(self):
        self.skip_unless_fl2008()

        # Test constrained delegation when the client service ticket does not
        # contain a PAC, and an empty msDS-AllowedToDelegateTo attribute.
        self._run_delegation_test(
            {
                'expected_error_mode': (KDC_ERR_MODIFIED,
                                        KDC_ERR_TGT_REVOKED),
                # We aren’t particular about whether or not we get an NTSTATUS.
                'expect_status': None,
                'expected_status': ntstatus.NT_STATUS_NOT_SUPPORTED,
                'allow_rbcd': True,
                'pac_options': '0001',  # supports RBCD
                'modify_client_tkt_fn': self.remove_ticket_pac,
                'service2_opts': {
                    'no_auth_data_required': True
                }
            })

    def test_rbcd_no_client_pac_no_auth_data_required_b(self):
        self.skip_unless_fl2008()

        # Test constrained delegation when the client service ticket does not
        # contain a PAC, and a non-empty msDS-AllowedToDelegateTo attribute.
        self._run_delegation_test(
            {
                'expected_error_mode': (KDC_ERR_MODIFIED,
                                        KDC_ERR_TGT_REVOKED),
                # We aren’t particular about whether or not we get an NTSTATUS.
                'expect_status': None,
                'expected_status': ntstatus.NT_STATUS_NO_MATCH,
                'allow_rbcd': True,
                'pac_options': '0001',  # supports RBCD
                'modify_client_tkt_fn': self.remove_ticket_pac,
                'service1_opts': {
                    'delegation_to_spn': ('host/test')
                },
                'service2_opts': {
                    'no_auth_data_required': True
                }
            })

    def test_rbcd_no_service_pac_no_auth_data_required(self):
        self.skip_unless_fl2008()

        # Test constrained delegation when the service TGT does not contain a
        # PAC.
        self._run_delegation_test(
            {
                'expected_error_mode': KDC_ERR_TGT_REVOKED,
                'allow_rbcd': True,
                'pac_options': '0001',  # supports RBCD
                'modify_service_tgt_fn': self.remove_ticket_pac,
                'service2_opts': {
                    'no_auth_data_required': True
                },
                'expect_edata': False
            })

    def test_rbcd_non_forwardable(self):
        self.skip_unless_fl2008()

        # Test resource-based constrained delegation with a non-forwardable
        # ticket.
        self._run_delegation_test(
            {
                'expected_error_mode': KDC_ERR_BADOPTION,
                # We aren’t particular about whether or not we get an NTSTATUS.
                'expect_status': None,
                'expected_status': ntstatus.NT_STATUS_ACCOUNT_RESTRICTION,
                'allow_rbcd': True,
                'pac_options': '0001',  # supports RBCD
                'modify_client_tkt_fn': functools.partial(
                    self.set_ticket_forwardable, flag=False)
            })

    def test_rbcd_no_pac_options_a(self):
        self.skip_unless_fl2008()

        # Test resource-based constrained delegation without the RBCD bit set
        # in the PAC options, and an empty msDS-AllowedToDelegateTo attribute.
        self._run_delegation_test(
            {
                'expected_error_mode': KDC_ERR_BADOPTION,
                # We aren’t particular about whether or not we get an NTSTATUS.
                'expect_status': None,
                'expected_status': ntstatus.NT_STATUS_NOT_SUPPORTED,
                'allow_rbcd': True,
                'pac_options': '1'  # does not support RBCD
            })

    def test_rbcd_no_pac_options_b(self):
        self.skip_unless_fl2008()

        # Test resource-based constrained delegation without the RBCD bit set
        # in the PAC options, and a non-empty msDS-AllowedToDelegateTo
        # attribute.
        self._run_delegation_test(
            {
                'expected_error_mode': KDC_ERR_BADOPTION,
                # We aren’t particular about whether or not we get an NTSTATUS.
                'expect_status': None,
                'expected_status': ntstatus.NT_STATUS_NO_MATCH,
                'allow_rbcd': True,
                'pac_options': '1',  # does not support RBCD
                'service1_opts': {
                    'delegation_to_spn': ('host/test')
                }
            })

    def test_bronze_bit_constrained_delegation_old_checksum(self):
        # Attempt to modify the ticket without updating the PAC checksums.
        self._run_delegation_test(
            {
                'expected_error_mode': (KDC_ERR_MODIFIED,
                                        KDC_ERR_BAD_INTEGRITY),
                'allow_delegation': True,
                'client_tkt_options': '0',  # non-forwardable ticket
                'modify_client_tkt_fn': functools.partial(
                    self.set_ticket_forwardable,
                    flag=True, update_pac_checksums=False),
                'expect_edata': False
            })

    def test_bronze_bit_rbcd_old_checksum(self):
        self.skip_unless_fl2008()

        # Attempt to modify the ticket without updating the PAC checksums.
        self._run_delegation_test(
            {
                'expected_error_mode': (KDC_ERR_MODIFIED,
                                        KDC_ERR_BAD_INTEGRITY),
                # We aren’t particular about whether or not we get an NTSTATUS.
                'expect_status': None,
                'expected_status': ntstatus.NT_STATUS_NOT_SUPPORTED,
                'allow_rbcd': True,
                'pac_options': '0001',  # supports RBCD
                'client_tkt_options': '0',  # non-forwardable ticket
                'modify_client_tkt_fn': functools.partial(
                    self.set_ticket_forwardable,
                    flag=True, update_pac_checksums=False)
            })

    def test_constrained_delegation_missing_client_checksum(self):
        # Present a user ticket without the required checksums.
        for checksum in self.pac_checksum_types:
            with self.subTest(checksum=checksum):
                if checksum == krb5pac.PAC_TYPE_TICKET_CHECKSUM:
                    expected_error_mode = (KDC_ERR_MODIFIED,
                                           KDC_ERR_BADOPTION)
                else:
                    expected_error_mode = KDC_ERR_GENERIC

                self._run_delegation_test(
                    {
                        'expected_error_mode': expected_error_mode,
                        'allow_delegation': True,
                        'modify_client_tkt_fn': functools.partial(
                            self.remove_pac_checksum, checksum=checksum),
                        'expect_edata': False
                    })

    def test_constrained_delegation_missing_service_checksum(self):
        # Present the service's ticket without the required checksums.
        for checksum in (krb5pac.PAC_TYPE_SRV_CHECKSUM,
                         krb5pac.PAC_TYPE_KDC_CHECKSUM):
            with self.subTest(checksum=checksum):
                self._run_delegation_test(
                    {
                        'expected_error_mode': KDC_ERR_GENERIC,
                        # We aren’t particular about whether or not we get an
                        # NTSTATUS.
                        'expect_status': None,
                        'expected_status':
                            ntstatus.NT_STATUS_INSUFFICIENT_RESOURCES,
                        'allow_delegation': True,
                        'modify_service_tgt_fn': functools.partial(
                            self.remove_pac_checksum, checksum=checksum)
                    })

    def test_rbcd_missing_client_checksum(self):
        self.skip_unless_fl2008()

        # Present a user ticket without the required checksums.
        for checksum in self.pac_checksum_types:
            with self.subTest(checksum=checksum):
                if checksum == krb5pac.PAC_TYPE_TICKET_CHECKSUM:
                    expected_error_mode = (KDC_ERR_MODIFIED,
                                           KDC_ERR_BADOPTION)
                else:
                    expected_error_mode = KDC_ERR_GENERIC

                self._run_delegation_test(
                    {
                        'expected_error_mode': expected_error_mode,
                        # We aren’t particular about whether or not we get an
                        # NTSTATUS.
                        'expect_status': None,
                        'expected_status':
                            ntstatus.NT_STATUS_NOT_SUPPORTED,
                        'allow_rbcd': True,
                        'pac_options': '0001',  # supports RBCD
                        'modify_client_tkt_fn': functools.partial(
                            self.remove_pac_checksum, checksum=checksum)
                    })

    def test_rbcd_missing_service_checksum(self):
        self.skip_unless_fl2008()

        # Present the service's ticket without the required checksums.
        for checksum in (krb5pac.PAC_TYPE_SRV_CHECKSUM,
                         krb5pac.PAC_TYPE_KDC_CHECKSUM):
            with self.subTest(checksum=checksum):
                self._run_delegation_test(
                    {
                        'expected_error_mode': KDC_ERR_GENERIC,
                        # We aren’t particular about whether or not we get an
                        # NTSTATUS.
                        'expect_status': None,
                        'expected_status':
                            ntstatus.NT_STATUS_INSUFFICIENT_RESOURCES,
                        'allow_rbcd': True,
                        'pac_options': '0001',  # supports RBCD
                        'modify_service_tgt_fn': functools.partial(
                            self.remove_pac_checksum, checksum=checksum)
                    })

    def test_constrained_delegation_zeroed_client_checksum(self):
        # Present a user ticket with invalid checksums.
        for checksum in self.pac_checksum_types:
            with self.subTest(checksum=checksum):
                self._run_delegation_test(
                    {
                        'expected_error_mode': (KDC_ERR_MODIFIED,
                                                KDC_ERR_BAD_INTEGRITY),
                        'allow_delegation': True,
                        'modify_client_tkt_fn': functools.partial(
                            self.zeroed_pac_checksum, checksum=checksum),
                        'expect_edata': False
                    })

    def test_constrained_delegation_zeroed_service_checksum(self):
        # Present the service's ticket with invalid checksums.
        for checksum in self.pac_checksum_types:
            with self.subTest(checksum=checksum):
                if checksum == krb5pac.PAC_TYPE_SRV_CHECKSUM:
                    expected_error_mode = (KDC_ERR_MODIFIED,
                                           KDC_ERR_BAD_INTEGRITY)
                    # We aren’t particular about whether or not we get an
                    # NTSTATUS.
                    expect_status = None
                    expected_status = ntstatus.NT_STATUS_WRONG_PASSWORD
                else:
                    expected_error_mode = 0
                    expect_status = None
                    expected_status = None

                self._run_delegation_test(
                    {
                        'expected_error_mode': expected_error_mode,
                        'expect_status': expect_status,
                        'expected_status': expected_status,
                        'allow_delegation': True,
                        'modify_service_tgt_fn': functools.partial(
                            self.zeroed_pac_checksum, checksum=checksum)
                    })

    def test_rbcd_zeroed_client_checksum(self):
        self.skip_unless_fl2008()

        # Present a user ticket with invalid checksums.
        for checksum in self.pac_checksum_types:
            with self.subTest(checksum=checksum):
                self._run_delegation_test(
                    {
                        'expected_error_mode': (KDC_ERR_MODIFIED,
                                                KDC_ERR_BAD_INTEGRITY),
                        # We aren’t particular about whether or not we get an
                        # NTSTATUS.
                        'expect_status': None,
                        'expected_status':
                            ntstatus.NT_STATUS_NOT_SUPPORTED,
                        'allow_rbcd': True,
                        'pac_options': '0001',  # supports RBCD
                        'modify_client_tkt_fn': functools.partial(
                            self.zeroed_pac_checksum, checksum=checksum)
                    })

    def test_rbcd_zeroed_service_checksum(self):
        self.skip_unless_fl2008()

        # Present the service's ticket with invalid checksums.
        for checksum in self.pac_checksum_types:
            with self.subTest(checksum=checksum):
                if checksum == krb5pac.PAC_TYPE_SRV_CHECKSUM:
                    expected_error_mode = (KDC_ERR_MODIFIED,
                                           KDC_ERR_BAD_INTEGRITY)
                    # We aren’t particular about whether or not we get an
                    # NTSTATUS.
                    expect_status = None
                    expected_status = ntstatus.NT_STATUS_WRONG_PASSWORD
                else:
                    expected_error_mode = 0
                    expect_status = None
                    expected_status = None

                self._run_delegation_test(
                    {
                        'expected_error_mode': expected_error_mode,
                        'expect_status': expect_status,
                        'expected_status': expected_status,
                        'allow_rbcd': True,
                        'pac_options': '0001',  # supports RBCD
                        'modify_service_tgt_fn': functools.partial(
                            self.zeroed_pac_checksum, checksum=checksum)
                    })

    unkeyed_ctypes = {Cksumtype.MD5, Cksumtype.SHA1, Cksumtype.CRC32}

    def test_constrained_delegation_unkeyed_client_checksum(self):
        # Present a user ticket with invalid checksums.
        for checksum in self.pac_checksum_types:
            for ctype in self.unkeyed_ctypes:
                with self.subTest(checksum=checksum, ctype=ctype):
                    if (checksum == krb5pac.PAC_TYPE_SRV_CHECKSUM
                            and ctype == Cksumtype.SHA1):
                        expected_error_mode = (KDC_ERR_SUMTYPE_NOSUPP,
                                               KDC_ERR_INAPP_CKSUM)
                    else:
                        expected_error_mode = (KDC_ERR_GENERIC,
                                               KDC_ERR_INAPP_CKSUM)

                    self._run_delegation_test(
                        {
                            'expected_error_mode': expected_error_mode,
                            'allow_delegation': True,
                            'modify_client_tkt_fn': functools.partial(
                                self.unkeyed_pac_checksum,
                                checksum=checksum, ctype=ctype),
                            'expect_edata': False
                        })

    def test_constrained_delegation_unkeyed_service_checksum(self):
        # Present the service's ticket with invalid checksums.
        for checksum in self.pac_checksum_types:
            for ctype in self.unkeyed_ctypes:
                with self.subTest(checksum=checksum, ctype=ctype):
                    if checksum == krb5pac.PAC_TYPE_SRV_CHECKSUM:
                        # We aren’t particular about whether or not we get an
                        # NTSTATUS.
                        expect_status = None
                        if ctype == Cksumtype.SHA1:
                            expected_error_mode = (KDC_ERR_SUMTYPE_NOSUPP,
                                                   KDC_ERR_INAPP_CKSUM)
                            expected_status = ntstatus.NT_STATUS_LOGON_FAILURE
                        else:
                            expected_error_mode = (KDC_ERR_GENERIC,
                                                   KDC_ERR_INAPP_CKSUM)
                            expected_status = (
                                ntstatus.NT_STATUS_INSUFFICIENT_RESOURCES)
                    else:
                        expected_error_mode = 0
                        expect_status = None
                        expected_status = None

                    self._run_delegation_test(
                        {
                            'expected_error_mode': expected_error_mode,
                            'expect_status': expect_status,
                            'expected_status': expected_status,
                            'allow_delegation': True,
                            'modify_service_tgt_fn': functools.partial(
                                self.unkeyed_pac_checksum,
                                checksum=checksum, ctype=ctype)
                        })

    def test_rbcd_unkeyed_client_checksum(self):
        self.skip_unless_fl2008()

        # Present a user ticket with invalid checksums.
        for checksum in self.pac_checksum_types:
            for ctype in self.unkeyed_ctypes:
                with self.subTest(checksum=checksum, ctype=ctype):
                    if (checksum == krb5pac.PAC_TYPE_SRV_CHECKSUM
                            and ctype == Cksumtype.SHA1):
                        expected_error_mode = (KDC_ERR_SUMTYPE_NOSUPP,
                                               KDC_ERR_INAPP_CKSUM)
                    else:
                        expected_error_mode = (KDC_ERR_GENERIC,
                                               KDC_ERR_INAPP_CKSUM)

                    self._run_delegation_test(
                        {
                            'expected_error_mode': expected_error_mode,
                            # We aren’t particular about whether or not we get
                            # an NTSTATUS.
                            'expect_status': None,
                            'expected_status':
                                ntstatus.NT_STATUS_NOT_SUPPORTED,
                            'allow_rbcd': True,
                            'pac_options': '0001',  # supports RBCD
                            'modify_client_tkt_fn': functools.partial(
                                self.unkeyed_pac_checksum,
                                checksum=checksum, ctype=ctype)
                        })

    def test_rbcd_unkeyed_service_checksum(self):
        self.skip_unless_fl2008()

        # Present the service's ticket with invalid checksums.
        for checksum in self.pac_checksum_types:
            for ctype in self.unkeyed_ctypes:
                with self.subTest(checksum=checksum, ctype=ctype):
                    if checksum == krb5pac.PAC_TYPE_SRV_CHECKSUM:
                        # We aren’t particular about whether or not we get an
                        # NTSTATUS.
                        expect_status = None
                        if ctype == Cksumtype.SHA1:
                            expected_error_mode = (KDC_ERR_SUMTYPE_NOSUPP,
                                                   KDC_ERR_INAPP_CKSUM)
                            expected_status = ntstatus.NT_STATUS_LOGON_FAILURE
                        else:
                            expected_error_mode = (KDC_ERR_GENERIC,
                                                   KDC_ERR_INAPP_CKSUM)
                            expected_status = (
                                ntstatus.NT_STATUS_INSUFFICIENT_RESOURCES)
                    else:
                        expected_error_mode = 0
                        expect_status = None
                        expected_status = None

                    self._run_delegation_test(
                        {
                            'expected_error_mode': expected_error_mode,
                            'expect_status': expect_status,
                            'expected_status': expected_status,
                            'allow_rbcd': True,
                            'pac_options': '0001',  # supports RBCD
                            'modify_service_tgt_fn': functools.partial(
                                self.unkeyed_pac_checksum,
                                checksum=checksum, ctype=ctype)
                        })

    def test_constrained_delegation_rc4_client_checksum(self):
        # Present a user ticket with RC4 checksums.
        samdb = self.get_samdb()
        functional_level = self.get_domain_functional_level(samdb)

        if functional_level >= dsdb.DS_DOMAIN_FUNCTION_2008:
            expected_error_mode = (KDC_ERR_GENERIC,
                                   KDC_ERR_INAPP_CKSUM)
            expect_edata = False
        else:
            expected_error_mode = 0
            expect_edata = None

        self._run_delegation_test(
            {
                'expected_error_mode': expected_error_mode,
                'allow_delegation': True,
                'modify_client_tkt_fn': self.rc4_pac_checksums,
                'expect_edata': expect_edata,
            })

    def test_rbcd_rc4_client_checksum(self):
        self.skip_unless_fl2008()

        # Present a user ticket with RC4 checksums.
        expected_error_mode = (KDC_ERR_GENERIC,
                               KDC_ERR_BADOPTION)

        self._run_delegation_test(
            {
                'expected_error_mode': expected_error_mode,
                # We aren’t particular about whether or not we get an NTSTATUS.
                'expect_status': None,
                'expected_status': ntstatus.NT_STATUS_NOT_SUPPORTED,
                'allow_rbcd': True,
                'pac_options': '0001',  # supports RBCD
                'modify_client_tkt_fn': self.rc4_pac_checksums,
            })

    def test_constrained_delegation_rodc_issued(self):
        self._run_delegation_test(
            {
                # Test that RODC-issued constrained delegation tickets are
                # accepted.
                'expected_error_mode': 0,
                'allow_delegation': True,
                # Both tickets must be signed by the same RODC.
                'modify_client_tkt_fn': self.signed_by_rodc,
                'modify_service_tgt_fn': self.issued_by_rodc,
                'client_opts': {
                    'allowed_replication_mock': True,
                    'revealed_to_mock_rodc': True,
                },
                'service1_opts': {
                    'allowed_replication_mock': True,
                    'revealed_to_mock_rodc': True,
                },
            })

    def test_rbcd_rodc_issued(self):
        self.skip_unless_fl2008()

        self._run_delegation_test(
            {
                # Test that RODC-issued constrained delegation tickets are
                # accepted.
                'expected_error_mode': 0,
                'allow_rbcd': True,
                'pac_options': '0001',  # supports RBCD
                # Both tickets must be signed by the same RODC.
                'modify_client_tkt_fn': self.signed_by_rodc,
                'modify_service_tgt_fn': self.issued_by_rodc,
                'client_opts': {
                    'allowed_replication_mock': True,
                    'revealed_to_mock_rodc': True,
                },
                'service1_opts': {
                    'allowed_replication_mock': True,
                    'revealed_to_mock_rodc': True,
                },
            })

    def remove_pac_checksum(self, ticket, checksum):
        checksum_keys = self.get_krbtgt_checksum_key()

        return self.modified_ticket(ticket,
                                    checksum_keys=checksum_keys,
                                    include_checksums={checksum: False})

    def zeroed_pac_checksum(self, ticket, checksum):
        krbtgt_creds = self.get_krbtgt_creds()
        krbtgt_key = self.TicketDecryptionKey_from_creds(krbtgt_creds)

        server_key = ticket.decryption_key

        checksum_keys = {
            krb5pac.PAC_TYPE_SRV_CHECKSUM: server_key,
            krb5pac.PAC_TYPE_KDC_CHECKSUM: krbtgt_key,
            krb5pac.PAC_TYPE_TICKET_CHECKSUM: krbtgt_key,
        }

        if checksum == krb5pac.PAC_TYPE_SRV_CHECKSUM:
            zeroed_key = server_key
        else:
            zeroed_key = krbtgt_key

        checksum_keys[checksum] = ZeroedChecksumKey(zeroed_key.key,
                                                    zeroed_key.kvno)

        return self.modified_ticket(ticket,
                                    checksum_keys=checksum_keys,
                                    include_checksums={checksum: True})

    def unkeyed_pac_checksum(self, ticket, checksum, ctype):
        krbtgt_creds = self.get_krbtgt_creds()
        krbtgt_key = self.TicketDecryptionKey_from_creds(krbtgt_creds)

        server_key = ticket.decryption_key

        checksum_keys = {
            krb5pac.PAC_TYPE_SRV_CHECKSUM: server_key,
            krb5pac.PAC_TYPE_KDC_CHECKSUM: krbtgt_key,
            krb5pac.PAC_TYPE_TICKET_CHECKSUM: krbtgt_key,
            krb5pac.PAC_TYPE_FULL_CHECKSUM: krbtgt_key,
        }

        # Make a copy of the existing key and change the ctype.
        key = checksum_keys[checksum]
        new_key = RodcPacEncryptionKey(key.key, key.kvno)
        new_key.ctype = ctype
        checksum_keys[checksum] = new_key

        return self.modified_ticket(ticket,
                                    checksum_keys=checksum_keys,
                                    include_checksums={checksum: True})

    def rc4_pac_checksums(self, ticket):
        krbtgt_creds = self.get_krbtgt_creds()
        rc4_krbtgt_key = self.TicketDecryptionKey_from_creds(
            krbtgt_creds, etype=Enctype.RC4)

        server_key = ticket.decryption_key

        checksum_keys = {
            krb5pac.PAC_TYPE_SRV_CHECKSUM: server_key,
            krb5pac.PAC_TYPE_KDC_CHECKSUM: rc4_krbtgt_key,
            krb5pac.PAC_TYPE_TICKET_CHECKSUM: rc4_krbtgt_key,
            krb5pac.PAC_TYPE_FULL_CHECKSUM: rc4_krbtgt_key,
        }

        include_checksums = {
            krb5pac.PAC_TYPE_SRV_CHECKSUM: True,
            krb5pac.PAC_TYPE_KDC_CHECKSUM: True,
            krb5pac.PAC_TYPE_TICKET_CHECKSUM: True,
            krb5pac.PAC_TYPE_FULL_CHECKSUM: True,
        }

        return self.modified_ticket(ticket,
                                    checksum_keys=checksum_keys,
                                    include_checksums=include_checksums)

    def add_delegation_info(self, ticket, *, services):
        def modify_pac_fn(pac):
            pac_buffers = pac.buffers
            self.assertNotIn(krb5pac.PAC_TYPE_CONSTRAINED_DELEGATION,
                             (buffer.type for buffer in pac_buffers))

            transited_services = list(map(lsa.String, services))

            delegation = krb5pac.PAC_CONSTRAINED_DELEGATION()
            delegation.proxy_target = lsa.String('test_proxy_target')
            delegation.transited_services = transited_services
            delegation.num_transited_services = len(transited_services)

            info = krb5pac.PAC_CONSTRAINED_DELEGATION_CTR()
            info.info = delegation

            pac_buffer = krb5pac.PAC_BUFFER()
            pac_buffer.type = krb5pac.PAC_TYPE_CONSTRAINED_DELEGATION
            pac_buffer.info = info

            pac_buffers.append(pac_buffer)

            pac.buffers = pac_buffers
            pac.num_buffers += 1

            return pac

        checksum_keys = self.get_krbtgt_checksum_key()

        return self.modified_ticket(ticket,
                                    checksum_keys=checksum_keys,
                                    modify_pac_fn=modify_pac_fn)

    def set_ticket_forwardable(self, ticket, flag, update_pac_checksums=True):
        modify_fn = functools.partial(self.modify_ticket_flag,
                                      flag='forwardable',
                                      value=flag)

        if update_pac_checksums:
            checksum_keys = self.get_krbtgt_checksum_key()
        else:
            checksum_keys = None

        return self.modified_ticket(ticket,
                                    modify_fn=modify_fn,
                                    checksum_keys=checksum_keys,
                                    update_pac_checksums=update_pac_checksums)

    def remove_ticket_pac(self, ticket):
        return self.modified_ticket(ticket,
                                    exclude_pac=True)


if __name__ == "__main__":
    global_asn1_print = False
    global_hexdump = False
    import unittest
    unittest.main()



# Unix SMB/CIFS implementation.
# Copyright (C) Catalyst IT Ltd. 2019
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from samba.tests.samba_tool.base import SambaToolCmdTest
import os
import shutil


class PromoteDcLmdbSizeTestCase(SambaToolCmdTest):
    """Test setting of the lmdb map size during a promote dc"""

    def setUp(self):
        super().setUp()
        self.tempsambadir = os.path.join(self.tempdir, "samba")
        os.mkdir(self.tempsambadir)
        (_, name) = os.path.split(self.tempdir)
        self.netbios_name = name

    # join a domain as a member server
    #
    # returns the tuple (ret, stdout, stderr)
    def join_member(self):
        command = (
            "samba-tool " +
            "domain join " +
            os.environ["REALM"] + " " +
            "member " +
            ("-U%s%%%s " % (os.environ["USERNAME"], os.environ["PASSWORD"])) +
            ("--option=netbiosname=%s " % self.netbios_name) +
            ("--targetdir=%s " % self.tempsambadir))
        return self.run_command(command)

    #
    # Promote a member server to a domain controller
    def promote(self, size=None, role=None):
        command = (
            "samba-tool " +
            "domain dcpromo " +
            os.environ["REALM"] + " " +
            role + " " +
            ("-U%s%%%s " % (os.environ["USERNAME"], os.environ["PASSWORD"])) +
            ("--option=netbiosname=%s " % self.netbios_name) +
            ("--targetdir=%s " % self.tempsambadir) +
            "--backend-store=mdb "
        )
        if size:
            command += ("--backend-store-size=%s" % size)

        (ret, stdout, stderr) = self.run_command(command)
        if ret == 0:
            self.cleanup_join(self.netbios_name)

        return (ret, stdout, stderr)

    def is_rodc(self):
        url = "ldb://%s/private/sam.ldb" % self.tempsambadir
        samdb = self.getSamDB("-H", url)
        return samdb.am_rodc()

    #
    # Get the lmdb map size for the specified command
    #
    # While there is a python lmdb package available we use the lmdb command
    # line utilities to avoid introducing a dependency.
    #
    def get_lmdb_environment_size(self, path):
        (result, out, err) = self.run_command("mdb_stat -ne %s" % path)
        if result:
            self.fail("Unable to run mdb_stat\n")
        for line in out.split("\n"):
            line = line.strip()
            if line.startswith("Map size:"):
                line = line.replace(" ", "")
                (label, size) = line.split(":")
                return int(size)

    #
    # Check the lmdb files created by join and ensure that the map size
    # has been set to size.
    #
    # Currently this is all the *.ldb files in private/sam.ldb.d
    #
    def check_lmdb_environment_sizes(self, size):
        directory = os.path.join(self.tempsambadir, "private", "sam.ldb.d")
        for name in os.listdir(directory):
            if name.endswith(".ldb"):
                path = os.path.join(directory, name)
                s = self.get_lmdb_environment_size(path)
                if s != size:
                    self.fail("File %s, size=%d larger than %d" %
                              (name, s, size))

    #
    # Ensure that if --backend-store-size is not specified the default of
    # 8Gb is used
    def test_promote_dc_default(self):
        (result, out, err) = self.join_member()
        self.assertEqual(0, result)
        (result, out, err) = self.promote(role="DC")
        self.assertEqual(0, result)
        self.check_lmdb_environment_sizes(8 * 1024 * 1024 * 1024)
        self.assertFalse(self.is_rodc())

    #
    # Ensure that if --backend-store-size is not specified the default of
    # 8Gb is used
    def test_promote_rodc_default(self):
        (result, out, err) = self.join_member()
        self.assertEqual(0, result)
        (result, out, err) = self.promote(role="RODC")
        self.assertEqual(0, result)
        self.check_lmdb_environment_sizes(8 * 1024 * 1024 * 1024)
        self.assertTrue(self.is_rodc())

    #
    # Promote to a DC with a backend size of 96Mb
    def test_promote_dc_96Mb(self):
        (result, out, err) = self.join_member()
        self.assertEqual(0, result)
        (result, out, err) = self.promote(role="DC", size="96Mb")
        self.assertEqual(0, result)
        self.check_lmdb_environment_sizes(96 * 1024 * 1024)
        self.assertFalse(self.is_rodc())

    #
    # Promote to an RODC with a backend size of 256Mb
    def test_promote_rodc_256Mb(self):
        (result, out, err) = self.join_member()
        self.assertEqual(0, result)
        (result, out, err) = self.promote(role="RODC", size="256Mb")
        self.assertEqual(0, result)
        self.check_lmdb_environment_sizes(256 * 1024 * 1024)
        self.assertTrue(self.is_rodc())

    def test_no_unit_suffix(self):
        (result, out, err) = self.run_command(
            'samba-tool domain dcpromo --backend-store-size "2"')
        self.assertGreater(result, 0)
        self.assertRegex(err,
                         r"--backend-store-size invalid suffix ''")

    def test_invalid_unit_suffix(self):
        (result, out, err) = self.run_command(
            'samba-tool domain dcpromo --backend-store-size "2 cd"')
        self.assertGreater(result, 0)
        self.assertRegex(err,
                         r"--backend-store-size invalid suffix 'cd'")

    def test_non_numeric(self):
        (result, out, err) = self.run_command(
            'samba-tool domain dcpromo --backend-store-size "two Gb"')
        self.assertGreater(result, 0)
        self.assertRegex(
            err,
            r"backend-store-size option requires a numeric value, with an"
            " optional unit suffix")

    def tearDown(self):
        super().tearDown()
        shutil.rmtree(self.tempsambadir)



# gp_sec_ext kdc gpo policy
# Copyright (C) Luke Morrison <luc785@.hotmail.com> 2013
# Copyright (C) David Mulder <dmulder@suse.com> 2018
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os.path
from samba.gp.gpclass import gp_inf_ext
from samba.auth import system_session
from samba.common import get_string
try:
    from ldb import LdbError
    from samba.samdb import SamDB
except ImportError:
    pass
from samba.gp.util.logging import log

def mins_to_hours(val):
    return '%d' % (int(val) / 60)

def days_to_hours(val):
    return '%d' % (int(val) * 24)

def days2rel_nttime(val):
    seconds = 60
    minutes = 60
    hours = 24
    sam_add = 10000000
    val = int(val)
    return str(-(val * seconds * minutes * hours * sam_add))

class gp_krb_ext(gp_inf_ext):
    apply_map = { 'MaxTicketAge':  'kdc:user_ticket_lifetime',
                  'MaxServiceAge': 'kdc:service_ticket_lifetime',
                  'MaxRenewAge':   'kdc:renewal_lifetime' }
    def process_group_policy(self, deleted_gpo_list, changed_gpo_list):
        if self.lp.get('server role') != 'active directory domain controller':
            return
        inf_file = 'MACHINE/Microsoft/Windows NT/SecEdit/GptTmpl.inf'
        for guid, settings in deleted_gpo_list:
            self.gp_db.set_guid(guid)
            for section in settings.keys():
                if section == str(self):
                    for att, value in settings[section].items():
                        self.set_kdc_tdb(att, value)
                        self.gp_db.delete(section, att)
                        self.gp_db.commit()

        for gpo in changed_gpo_list:
            if gpo.file_sys_path:
                self.gp_db.set_guid(gpo.name)
                path = os.path.join(gpo.file_sys_path, inf_file)
                inf_conf = self.parse(path)
                if not inf_conf:
                    continue
                for section in inf_conf.sections():
                    if section == str(self):
                        for key, value in inf_conf.items(section):
                            if key not in gp_krb_ext.apply_map:
                                continue
                            att = gp_krb_ext.apply_map[key]
                            value_func = self.mapper().get(att)
                            self.set_kdc_tdb(att, value_func(value))
                            self.gp_db.commit()

    def set_kdc_tdb(self, attribute, val):
        old_val = self.gp_db.gpostore.get(attribute)
        log.info('%s was changed from %s to %s' % (attribute, old_val, val))
        if val is not None:
            self.gp_db.gpostore.store(attribute, get_string(val))
            self.gp_db.store(str(self), attribute, get_string(old_val)
                    if old_val else None)
        else:
            self.gp_db.gpostore.delete(attribute)
            self.gp_db.delete(str(self), attribute)

    def mapper(self):
        return {'kdc:user_ticket_lifetime': lambda val: val,
                'kdc:service_ticket_lifetime': mins_to_hours,
                'kdc:renewal_lifetime': days_to_hours,
                }

    def __str__(self):
        return 'Kerberos Policy'

    def rsop(self, gpo):
        output = {}
        if self.lp.get('server role') != 'active directory domain controller':
            return output
        inf_file = 'MACHINE/Microsoft/Windows NT/SecEdit/GptTmpl.inf'
        if gpo.file_sys_path:
            path = os.path.join(gpo.file_sys_path, inf_file)
            inf_conf = self.parse(path)
            if not inf_conf:
                return output
            if str(self) in inf_conf.sections():
                section = str(self)
                output[section] = {k: v for k, v in inf_conf.items(section)
                                      if gp_krb_ext.apply_map.get(k)}
        return output


class gp_access_ext(gp_inf_ext):
    """This class takes the .inf file parameter (essentially a GPO file mapped
    to a GUID), hashmaps it to the Samba parameter, which then uses an ldb
    object to update the parameter to Samba4. Not registry oriented whatsoever.
    """

    def load_ldb(self):
        try:
            self.ldb = SamDB(self.lp.samdb_url(),
                             session_info=system_session(),
                             credentials=self.creds,
                             lp=self.lp)
        except (NameError, LdbError):
            raise Exception('Failed to load SamDB for assigning Group Policy')

    apply_map = { 'MinimumPasswordAge':     'minPwdAge',
                  'MaximumPasswordAge':     'maxPwdAge',
                  'MinimumPasswordLength':  'minPwdLength',
                  'PasswordComplexity':     'pwdProperties' }
    def process_group_policy(self, deleted_gpo_list, changed_gpo_list):
        if self.lp.get('server role') != 'active directory domain controller':
            return
        self.load_ldb()
        inf_file = 'MACHINE/Microsoft/Windows NT/SecEdit/GptTmpl.inf'
        for guid, settings in deleted_gpo_list:
            self.gp_db.set_guid(guid)
            for section in settings.keys():
                if section == str(self):
                    for att, value in settings[section].items():
                        update_samba, _ = self.mapper().get(att)
                        update_samba(att, value)
                        self.gp_db.delete(section, att)
                        self.gp_db.commit()

        for gpo in changed_gpo_list:
            if gpo.file_sys_path:
                self.gp_db.set_guid(gpo.name)
                path = os.path.join(gpo.file_sys_path, inf_file)
                inf_conf = self.parse(path)
                if not inf_conf:
                    continue
                for section in inf_conf.sections():
                    if section == str(self):
                        for key, value in inf_conf.items(section):
                            if key not in gp_access_ext.apply_map:
                                continue
                            att = gp_access_ext.apply_map[key]
                            (update_samba, value_func) = self.mapper().get(att)
                            update_samba(att, value_func(value))
                            self.gp_db.commit()

    def ch_minPwdAge(self, attribute, val):
        old_val = self.ldb.get_minPwdAge()
        log.info('KDC Minimum Password age was changed from %s to %s'
                 % (old_val, val))
        self.gp_db.store(str(self), attribute, str(old_val))
        self.ldb.set_minPwdAge(val)

    def ch_maxPwdAge(self, attribute, val):
        old_val = self.ldb.get_maxPwdAge()
        log.info('KDC Maximum Password age was changed from %s to %s'
                 % (old_val, val))
        self.gp_db.store(str(self), attribute, str(old_val))
        self.ldb.set_maxPwdAge(val)

    def ch_minPwdLength(self, attribute, val):
        old_val = self.ldb.get_minPwdLength()
        log.info('KDC Minimum Password length was changed from %s to %s'
                 % (old_val, val))
        self.gp_db.store(str(self), attribute, str(old_val))
        self.ldb.set_minPwdLength(val)

    def ch_pwdProperties(self, attribute, val):
        old_val = self.ldb.get_pwdProperties()
        log.info('KDC Password Properties were changed from %s to %s'
                  % (old_val, val))
        self.gp_db.store(str(self), attribute, str(old_val))
        self.ldb.set_pwdProperties(val)

    def mapper(self):
        """ldap value : samba setter"""
        return {"minPwdAge": (self.ch_minPwdAge, days2rel_nttime),
                "maxPwdAge": (self.ch_maxPwdAge, days2rel_nttime),
                # Could be none, but I like the method assignment in
                # update_samba
                "minPwdLength": (self.ch_minPwdLength, lambda val: val),
                "pwdProperties": (self.ch_pwdProperties, lambda val: val),

                }

    def __str__(self):
        return 'System Access'

    def rsop(self, gpo):
        output = {}
        if self.lp.get('server role') != 'active directory domain controller':
            return output
        inf_file = 'MACHINE/Microsoft/Windows NT/SecEdit/GptTmpl.inf'
        if gpo.file_sys_path:
            path = os.path.join(gpo.file_sys_path, inf_file)
            inf_conf = self.parse(path)
            if not inf_conf:
                return output
            if str(self) in inf_conf.sections():
                section = str(self)
                output[section] = {k: v for k, v in inf_conf.items(section)
                                      if gp_access_ext.apply_map.get(k)}
        return output



# Unix SMB/CIFS implementation.
# Copyright © Catalyst
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from samba.tests import TestCase
from samba import compression


TEST_DIR = "testdata/compression"


class BaseCompressionTest(TestCase):
    def round_trip(self, data, size_delta=0):
        """Compress, decompress, assert equality with original.

        If size_delta is None, no size is given to decompress. This
        should fail with the Huffman variant and succeed with plain.
        Otherwise size_delta is added to the gven size; if negative,
        we'd expect a failure, with plain compression a positive delta
        will succeed.
        """

        compressed = self.compress(data)
        if size_delta is None:
            decompressed = self.decompress(compressed)
        else:
            decomp_size = len(data) + size_delta
            decompressed = self.decompress(compressed, decomp_size)

        if isinstance(data, str):
            data = data.encode()

        self.assertEqual(data, decompressed)
        return compressed

    def decompress_file(self, fn):
        decomp_fn = os.path.join(TEST_DIR,
                                 "decompressed",
                                 fn + ".decomp")
        comp_fn = os.path.join(TEST_DIR,
                               self.compressed_dir,
                               fn + self.compressed_suffix)

        with open(decomp_fn, 'rb') as f:
            decomp_expected = f.read()
        with open(comp_fn, 'rb') as f:
            comp = f.read()

        decompressed = self.decompress(comp, len(decomp_expected))

        self.assertEqual(decomp_expected, decompressed)


class LzxpressPlainCompressionTest(BaseCompressionTest):
    compress = compression.plain_compress
    decompress = compression.plain_decompress
    compressed_dir = "compressed-plain"
    compressed_suffix = ".lzplain"

    def test_round_trip_aaa_str(self):
        s = 'a' * 150000
        self.round_trip(s)

    def test_round_trip_aaa_bytes(self):
        s = b'a' * 150000
        self.round_trip(s)

    def test_round_trip_aaa_short(self):
        s = b'a' * 150000

        # this'll fail because the match for 'aaa...' will run
        # past the end of the buffer
        self.assertRaises(compression.CompressionError,
                          self.round_trip, s, -1)

    def test_round_trip_aaa_long(self):
        s = b'a' * 150000
        # this *won't* fail because although the data will run out
        # before the buffer is full, LZXpress plain does not care
        # about that.
        try:
            self.round_trip(s, 1)
        except compression.CompressionError as e:
            self.fail(f"failed to decompress with {e}")

    def test_round_trip_aaab_short(self):
        s = b'a' * 150000 + b'b'

        # this will *partially* succeed, because the buffer will fill
        # up vat a break in the decompression (not mid-match), and
        # lzxpress plain does not mind that. However self.round_trip
        # also makes an assertion that the original data equals the
        # decompressed result, and it won't because the decompressed
        # result is one byte shorter.
        self.assertRaises(AssertionError,
                          self.round_trip, s, -1)

    def test_round_trip_aaab_unstated(self):
        s = b'a' * 150000 + b'b'

        # this will succeed, because with no target size given, we
        # guess a large buffer in the python bindings.
        try:
            self.round_trip(s)
        except compression.CompressionError as e:
            self.fail(f"failed to decompress with {e}")

    def test_round_trip_30mb(self):
        s = b'abc' * 10000000
        # This will decompress into a string bigger than the python
        # bindings are willing to speculatively allocate, so will fail
        # to decompress.
        with self.assertRaises(compression.CompressionError):
            self.round_trip(s, None)

        # but it will be fine if we use the length
        try:
            self.round_trip(s, 0)
        except compression.CompressionError as e:
            self.fail(f"failed to decompress with {e}")

    def test_files(self):
        # We don't go through the whole set, which are already tested
        # by lib/compression/tests/test_lzxpress_plain.c
        for fn in ("slow-33d90a24e70515b14cd0",
                   "midsummer-nights-dream.txt"):
            self.decompress_file(fn)

    def test_empty_round_trip(self):
        # not symmetrical with Huffman, this doesn't fail
        self.round_trip('')


class LzxpressHuffmanCompressionTest(BaseCompressionTest):
    compress = compression.huffman_compress
    decompress = compression.huffman_decompress
    compressed_dir = "compressed-huffman"
    compressed_suffix = ".lzhuff"

    def test_round_trip_aaa_str(self):
        s = 'a' * 150000
        self.round_trip(s)

    def test_round_trip_aaa_bytes(self):
        s = b'a' * 150000
        self.round_trip(s)

    def test_round_trip_aaa_short(self):
        s = b'a' * 150000

        # this'll fail because the match for 'aaa...' will run
        # past the end of the buffer
        self.assertRaises(compression.CompressionError,
                          self.round_trip, s, -1)

    def test_round_trip_aaa_long(self):
        s = b'a' * 150000

        # this'll fail because the data will run out before the buffer
        # is full.
        self.assertRaises(compression.CompressionError,
                          self.round_trip, s, 1)

    def test_round_trip_aaab_short(self):
        s = b'a' * 150000 + b'b'

        # this *could* be allowed to succeed, because even though we
        # give it the wrong size, we know the decompression will not
        # flow over the end of the buffer, The behaviour here appears
        # to be implementation dependent -- the decompressor has the
        # option of saying 'whatever' and continuing. We are probably
        # stricter than Windows.
        self.assertRaises(compression.CompressionError,
                          self.round_trip, s, -1)

    def test_round_trip_aaab_unstated(self):
        s = b'a' * 150000 + b'b'

        # For the Huffman algorithm, the length is really an essential
        # part of the compression data, and the bindings will reject a
        # call with out it. This happens at the argument parsing stage,
        # so is a TypeError (i.e. wrong type of function), not a
        # CompressionError.
        self.assertRaises(TypeError,
                          self.round_trip, s, None)

    def test_files(self):
        # We don't go through the whole set, which are already tested
        # by lib/compression/tests/test_lzx_huffman.c
        for fn in ("slow-33d90a24e70515b14cd0",
                   "midsummer-nights-dream.txt"):
            self.decompress_file(fn)

    def test_empty_round_trip(self):
        with self.assertRaises(compression.CompressionError):
            self.round_trip('')
